#####################################################################
#                                                                   #
# /plugins/delete_repeated_shots/__init__.py                        #
#                                                                   #
# Copyright 2017, JQI                                               #
#                                                                   #
# This file is part of the program BLACS, in the labscript suite    #
# (see http://labscriptsuite.org), and is licensed under the        #
# Simplified BSD License. See the license.txt file in the root of   #
# the project for the full license.                                 #
#                                                                   #
#####################################################################
import logging
import os
import subprocess
import threading
import sys
from queue import Queue

from qtutils import UiLoader

from labscript_utils.shared_drive import path_to_agnostic
from labscript_utils.ls_zprocess import Lock
from blacs.plugins import PLUGINS_DIR

name = "Delete repeated shots"
module = "delete_repeated_shots" # should be folder name
logger = logging.getLogger('BLACS.plugin.%s'%module)


KEEP_ALL_SHOTS = 0

class Plugin(object):
    def __init__(self, initial_settings):
        self.menu = None
        self.notifications = {}
        self.initial_settings = initial_settings
        self.BLACS = None
        self.ui = None
        self.n_shots_to_keep = initial_settings.get('n_shots_to_keep', KEEP_ALL_SHOTS)
        self.delete_queue = initial_settings.get('delete_queue', [])
        self.event_queue = Queue()
        self.delete_queue_lock = threading.Lock()
        self.mainloop_thread = threading.Thread(target=self.mainloop)
        self.mainloop_thread.daemon = True
        self.mainloop_thread.start()
        
    def plugin_setup_complete(self, BLACS):
        self.BLACS = BLACS

        # Add our controls to the BLACS UI:
        self.ui = UiLoader().load(os.path.join(PLUGINS_DIR, module, 'controls.ui'))
        BLACS['ui'].queue_controls_frame.layout().addWidget(self.ui)

        # Restore settings to the GUI controls:
        self.ui.spinBox.setValue(self.n_shots_to_keep)

        # Connect signals:
        self.ui.spinBox.valueChanged.connect(self.on_spinbox_value_changed)
        self.ui.reset_button.clicked.connect(self.on_reset_button_clicked)
        BLACS['ui'].queue_repeat_button.toggled.connect(self.ui.setEnabled)

        # Our control is only enabled when repeat mode is active:
        self.ui.setEnabled(BLACS['ui'].queue_repeat_button.isChecked())

    def on_spinbox_value_changed(self, value):
        with self.delete_queue_lock:
            self.n_shots_to_keep = value
            # If the user reduces the number of shots to keep, but we had a
            # larger list of shots awaiting deletion, remove shots from the
            # deletion queue (without deleting them) until the queue is the
            # same size as the number of shots we are now keeping. This means
            # that if we set to keep 100 shots, and then we go ahead and run a
            # hundred shots, if we then set it to keep 5 shots it won't delete
            # the 95 oldest shots in the queue. Rather it will only delete the
            # most recent 5 (and not immediately - over the next 5 shots).
            while len(self.delete_queue) > self.n_shots_to_keep:
                self.delete_queue.pop(0)

    def on_reset_button_clicked(self):
        self.ui.spinBox.setValue(KEEP_ALL_SHOTS)

    def get_save_data(self):
        return {'n_shots_to_keep': self.n_shots_to_keep,
                'delete_queue': self.delete_queue}
    
    def get_callbacks(self):
        return {'shot_complete': self.on_shot_complete}
        
    def on_shot_complete(self, h5_filepath):

        # If we're keeping all shots, then there's nothing to do here:
        if self.n_shots_to_keep == KEEP_ALL_SHOTS:
            return

        # Is the file a repeated shot?
        basename, ext = os.path.splitext(os.path.basename(h5_filepath))
        if '_rep' in basename and ext == '.h5':
            repno = basename.split('_rep')[-1]
            try:
                int(repno)
            except ValueError:
                # not a rep:
                return
            else:
                # Yes, it is a rep. Queue it for deletion:
                self.delete_queue.append(h5_filepath)
                self.event_queue.put('shot complete')

    def mainloop(self):
        # We delete shots in a separate thread so that we don't slow down the queue waiting on
        # network communication to acquire the lock, 
        while True:
            try:
                event = self.event_queue.get()
                if event == 'close':
                    break
                elif event == 'shot complete':
                    while len(self.delete_queue) > self.n_shots_to_keep:
                        with self.delete_queue_lock:
                            h5_filepath = self.delete_queue.pop(0)
                        # Acquire a lock on the file so that we don't
                        # delete it whilst someone else has it open:
                        with Lock(path_to_agnostic(h5_filepath)):
                            try:
                                os.unlink(h5_filepath)
                                logger.info("Deleted repeated shot file %s" % h5_filepath)
                            except OSError:
                                logger.exception("Couldn't delete shot file %s" % h5_filepath)
                else:
                    raise ValueError(event)
            except Exception:
                logger.exception("Exception in repeated shot deletion loop, ignoring.")
    

    def close(self):
        self.event_queue.put('close')
        self.mainloop_thread.join()


    # The rest of these are boilerplate:
    def get_menu_class(self):
        return None
        
    def get_notification_classes(self):
        return []
        
    def get_setting_classes(self):
        return []
    
    def set_menu_instance(self, menu):
        self.menu = menu
        
    def set_notification_instances(self, notifications):
        self.notifications = notifications
        
    



import os
from pathlib import Path
try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

root = Path(__file__).parent.parent
if (root / '.git').is_dir():
    from setuptools_scm import get_version
    __version__ = get_version(
        root,
        version_scheme="release-branch-semver",
        local_scheme=os.getenv("SCM_LOCAL_SCHEME", "node-and-date"),
    )
else:
    try:
        __version__ = importlib_metadata.version(__package__)
    except importlib_metadata.PackageNotFoundError:
        __version__ = None



#####################################################################
#                                                                   #
# /analysis_submission.py                                           #
#                                                                   #
# Copyright 2013, Monash University                                 #
#                                                                   #
# This file is part of the program BLACS, in the labscript suite    #
# (see http://labscriptsuite.org), and is licensed under the        #
# Simplified BSD License. See the license.txt file in the root of   #
# the project for the full license.                                 #
#                                                                   #
#####################################################################
import logging
import os
import threading
import time
import sys
import queue

from qtutils.qt.QtCore import *
from qtutils.qt.QtGui import *
from qtutils.qt.QtWidgets import *

from qtutils import *
from zprocess import TimeoutError, raise_exception_in_thread
from zprocess.security import AuthenticationFailure
from labscript_utils.ls_zprocess import zmq_get
from socket import gaierror
import labscript_utils.shared_drive
from labscript_utils.qtwidgets.elide_label import elide_label
from blacs import BLACS_DIR


class AnalysisSubmission(object):        
    def __init__(self, BLACS, blacs_ui):
        self.inqueue = queue.Queue()
        self.BLACS = BLACS
        self.port = int(self.BLACS.exp_config.get('ports', 'lyse'))
        
        self._ui = UiLoader().load(os.path.join(BLACS_DIR, 'analysis_submission.ui'))
        blacs_ui.analysis.addWidget(self._ui)
        self._ui.frame.setMinimumWidth(blacs_ui.queue_controls_frame.sizeHint().width())
        elide_label(self._ui.resend_shots_label, self._ui.failed_to_send_frame.layout(), Qt.ElideRight)
        # connect signals
        self._ui.send_to_server.toggled.connect(lambda state: self._set_send_to_server(state))
        self._ui.server.editingFinished.connect(lambda: self._set_server(self._ui.server.text()))
        self._ui.clear_unsent_shots_button.clicked.connect(lambda _: self.clear_waiting_files())
        self._ui.retry_button.clicked.connect(lambda _: self.check_retry())

        self._waiting_for_submission = []
        self.failure_reason = None
        self.server_online = 'offline'
        self.send_to_server = False
        self.server = ''
        self.time_of_last_connectivity_check = 0

        self.mainloop_thread = threading.Thread(target=self.mainloop)
        self.mainloop_thread.daemon = True
        self.mainloop_thread.start()
        
    def restore_save_data(self,data):
        if "server" in data:
            self.server = data["server"]
        if "send_to_server" in data:
            self.send_to_server = data["send_to_server"]
        if "waiting_for_submission" in data:
            self._waiting_for_submission = list(data["waiting_for_submission"])
        self.inqueue.put(['save data restored', None])
        self.check_retry()
            
    def get_save_data(self):
        return {"waiting_for_submission":list(self._waiting_for_submission),
                "server":self.server,
                "send_to_server":self.send_to_server
               }
    
    def _set_send_to_server(self,value):
        self.send_to_server = value
        
    def _set_server(self,server):
        self.server = server
        self.check_retry()
    
    @property
    @inmain_decorator(True)
    def send_to_server(self):
        return self._send_to_server
        
    @send_to_server.setter
    @inmain_decorator(True)
    def send_to_server(self, value):
        self._send_to_server = bool(value)
        self._ui.send_to_server.setChecked(self.send_to_server)
        if self.send_to_server:
            self._ui.server.setEnabled(True)
            self._ui.server_online.show()
            self.check_retry()
        else:
            self.clear_waiting_files()
            self._ui.server.setEnabled(False)
            self._ui.server_online.hide()
    
    @property
    @inmain_decorator(True)
    def server(self):
        return str(self._server)
        
    @server.setter    
    @inmain_decorator(True)
    def server(self,value):
        self._server = value
        self._ui.server.setText(self.server)

    @property
    @inmain_decorator(True)
    def server_online(self):
        return self._server_online
        
    @server_online.setter
    @inmain_decorator(True)
    def server_online(self,value):
        self._server_online = str(value)
        
        icon_names = {'checking': ':/qtutils/fugue/hourglass',
                      'online': ':/qtutils/fugue/tick',
                      'offline': ':/qtutils/fugue/exclamation', 
                      '': ':/qtutils/fugue/status-offline'}

        tooltips = {'checking': 'Checking...',
                    'online': 'Server is responding',
                    'offline': 'Server not responding',
                    '': 'Disabled'}

        icon = QIcon(icon_names.get(self._server_online, ':/qtutils/fugue/exclamation-red'))
        pixmap = icon.pixmap(QSize(16, 16))
        tooltip = tooltips.get(self._server_online, "Invalid server status: %s" % self._server_online)
        if self.failure_reason is not None:
            tooltip += '\n' + self.failure_reason

        # Update GUI:
        self._ui.server_online.setPixmap(pixmap)
        self._ui.server_online.setToolTip(tooltip)
        self.update_waiting_files_message()


    @inmain_decorator(True)
    def update_waiting_files_message(self):
        # if there is only one shot and we haven't encountered failure yet, do
        # not show the error frame:
        if (self.server_online == 'checking') and (len(self._waiting_for_submission) == 1) and not self._ui.failed_to_send_frame.isVisible():
            return
        if self._waiting_for_submission:
            self._ui.failed_to_send_frame.show()
            if self.server_online == 'checking':
                self._ui.retry_button.hide()
                text = 'Sending %s shot(s)...' % len(self._waiting_for_submission)
            else:
                self._ui.retry_button.show()
                text = '%s shot(s) to send' % len(self._waiting_for_submission)
            self._ui.resend_shots_label.setText(text)
        else:
            self._ui.failed_to_send_frame.hide()

    def get_queue(self):
        return self.inqueue

    @inmain_decorator(True)
    def clear_waiting_files(self):
        self._waiting_for_submission = []
        self.update_waiting_files_message()

    @inmain_decorator(True)
    def check_retry(self):
        self.inqueue.put(['check/retry', None])

    def mainloop(self):
        self._mainloop_logger = logging.getLogger('BLACS.AnalysisSubmission.mainloop')
        # Ignore signals until save data is restored:
        while self.inqueue.get()[0] != 'save data restored':
            pass
        timeout = 10
        while True:
            try:
                try:
                    signal, data = self.inqueue.get(timeout=timeout)
                except queue.Empty:
                    timeout = 10
                    # Periodic checking of connectivity and resending of files.
                    # Don't trigger a re-check if we already failed a connectivity
                    # check within the last second:
                    if (time.time() - self.time_of_last_connectivity_check) > 1:
                        signal = 'check/retry'
                    else:
                        continue
                if signal == 'check/retry':
                    self.check_connectivity()
                    if self.server_online == 'online':
                        self.submit_waiting_files()
                elif signal == 'file':
                    if self.send_to_server:
                        self._waiting_for_submission.append(data)
                        if self.server_online != 'online':
                            # Don't stack connectivity checks if many files are
                            # arriving. If we failed a connectivity check less
                            # than a second ago then don't check again.
                            if (time.time() - self.time_of_last_connectivity_check) > 1:
                                self.check_connectivity()
                            else:
                                # But do queue up a check for when we have
                                # been idle for one second:
                                timeout = 1
                        if self.server_online == 'online':
                            self.submit_waiting_files()
                elif signal == 'close':
                    break
                elif signal == 'save data restored':
                    continue
                else:
                    raise ValueError('Invalid signal: %s'%str(signal))

                self._mainloop_logger.info('Processed signal: %s'%str(signal))
            except Exception:
                # Raise in a thread for visibility, but keep going
                raise_exception_in_thread(sys.exc_info())
                self._mainloop_logger.exception("Exception in mainloop, continuing")
            
    def check_connectivity(self):
        host = self.server
        send_to_server = self.send_to_server
        if host and send_to_server:       
            self.server_online = 'checking'         
            try:
                response = zmq_get(self.port, host, 'hello', timeout=1)
                self.failure_reason = None
            except (TimeoutError, gaierror, AuthenticationFailure) as e:
                success = False
                self.failure_reason = str(e)
            else:
                success = (response == 'hello')
                if not success:
                    self.failure_reason = "unexpected reponse: %s" % str(response)
                
            # update GUI
            self.server_online = 'online' if success else 'offline'
        else:
            self.server_online = ''

        self.time_of_last_connectivity_check = time.time()

    def submit_waiting_files(self):
        success = True
        while self._waiting_for_submission and success:
            path = self._waiting_for_submission[0]
            self._mainloop_logger.info('Submitting run file %s.\n'%os.path.basename(path))
            data = {'filepath': labscript_utils.shared_drive.path_to_agnostic(path)}
            self.server_online = 'checking'
            try:
                response = zmq_get(self.port, self.server, data, timeout=1)
                self.failure_reason = None
            except (TimeoutError, gaierror, AuthenticationFailure) as e:
                success = False
                self.failure_reason = str(e)
            else:
                success = (response == 'added successfully')
                if not success:
                    self.failure_reason = "unexpected reponse: %s" % str(response)
                try:
                    self._waiting_for_submission.pop(0) 
                except IndexError:
                    # Queue has been cleared
                    pass
            if not success:
                break
        # update GUI
        self.server_online = 'online' if success else 'offline'
        self.time_of_last_connectivity_check = time.time()
        



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import copy
import os
from pathlib import Path
from m2r import MdInclude
from recommonmark.transform import AutoStructify
from jinja2 import FileSystemLoader, Environment

# -- Project information (unique to each project) -------------------------------------

project = "blacs"
copyright = "2020, labscript suite"
author = "labscript suite contributors"

# The full version, including alpha/beta/rc tags
from blacs import __version__ as version  # noqa: E402

release = version

# HTML icons
img_path = 'img'
html_logo = img_path + "/blacs_64x64.svg"
html_favicon = img_path + "/blacs.ico"

# -- General configuration (should be identical across all projects) ------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_rtd_theme",
    "recommonmark",
]

autodoc_typehints = 'description'
autosummary_generate = True
numfig = True
autodoc_mock_imports = ['labscript_utils']

# mock missing site package methods
import site
mock_site_methods = {
    # Format:
    #   method name: return value
    'getusersitepackages': '',
    'getsitepackages': []
}
__fn = None
for __name, __rval in mock_site_methods.items():
    if not hasattr(site, __name):
        __fn = lambda *args, __rval=copy.deepcopy(__rval), **kwargs: __rval
        setattr(site, __name, __fn)
del __name
del __rval
del __fn

# Prefix each autosectionlabel with the name of the document it is in and a colon
autosectionlabel_prefix_document = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The suffix(es) of source filenames.
source_suffix = ['.rst', '.md']

# The master toctree document.
master_doc = 'index'

# intersphinx allows us to link directly to other repos sphinxdocs.
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
    'qtutils': ('https://qtutils.readthedocs.io/en/stable/', None),
    'pyqtgraph': (
        'https://pyqtgraph.readthedocs.io/en/latest/',
        None,
    ),  # change to stable once v0.11 is published
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'h5py': ('https://docs.h5py.org/en/stable/', None),
    'pydaqmx': ('https://pythonhosted.org/PyDAQmx/', None),
    'qt': (
        '',
        'pyqt5-modified-objects.inv',
    )  # from https://github.com/MSLNZ/msl-qt/blob/master/docs/create_pyqt_objects.py
    # under MIT License
    # TODO
    # desktop-app
    # spinapi/pynivision/etc
}

# list of all labscript suite components that have docs
labscript_suite_programs = {
    'labscript': {
        'desc': 'Expressive composition of hardware-timed experiments',
        'icon': 'labscript_32nx32n.svg',
        'type': 'lib',
    },
    'labscript-devices': {
        'desc': 'Plugin architecture for controlling experiment hardware',
        'icon': 'labscript_32nx32n.svg',
        'type': 'lib',
    },
    'labscript-utils': {
        'desc': 'Shared modules used by the *labscript suite*',
        'icon': 'labscript_32nx32n.svg',
        'type': 'lib',
    },
    'runmanager': {
        'desc': 'Graphical and remote interface to parameterized experiments',
        'icon': 'runmanager_32nx32n.svg',
        'type': 'gui',
    },
    'blacs': {
        'desc': 'Graphical interface to scientific instruments and experiment supervision',
        'icon': 'blacs_32nx32n.svg',
        'type': 'gui',
    },
    'lyse': {
        'desc': 'Online analysis of live experiment data',
        'icon': 'lyse_32nx32n.svg',
        'type': 'gui',
    },
    'runviewer': {
        'desc': 'Visualize hardware-timed experiment instructions',
        'icon': 'runviewer_32nx32n.svg',
        'type': 'gui',
    },
}

# whether to use stable or latest version
labscript_suite_doc_version = os.environ.get('READTHEDOCS_VERSION', 'latest')
if '.' in labscript_suite_doc_version:
    labscript_suite_doc_version = 'stable'
elif labscript_suite_doc_version not in ['stable', 'latest']:
    labscript_suite_doc_version = 'latest'

# add intersphinx references for each component
for ls_prog in labscript_suite_programs:
    intersphinx_mapping[ls_prog] = (
        'https://docs.labscriptsuite.org/projects/{}/en/{}/'.format(
            ls_prog, labscript_suite_doc_version
        ),
        None,
    )

# add intersphinx reference for the metapackage
if project != "the labscript suite":
    intersphinx_mapping['labscript-suite'] = (
        'https://docs.labscriptsuite.org/en/{}/'.format(labscript_suite_doc_version),
        None,
    )

# Make `some code` equivalent to :code:`some code`
default_role = 'code'

# hide todo notes if on readthedocs and not building the latest
if os.environ.get('READTHEDOCS') and (
    os.environ.get('READTHEDOCS_VERSION') != 'latest'
    or (
        os.environ.get('READTHEDOCS_PROJECT') == project
        or os.environ.get('READTHEDOCS_PROJECT') == 'labscriptsuite'
    )
):
    todo_include_todos = False
else:
    todo_include_todos = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_title = "labscript suite | {project}".format(
    project=project
    if project != "labscript-suite"
    else "experiment control and automation"
)
html_short_title = "labscript suite"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Customize the html_theme
html_theme_options = {'navigation_depth': 3}

# Use m2r only for mdinclude and recommonmark for everything else
# https://github.com/readthedocs/recommonmark/issues/191#issuecomment-622369992
def setup(app):
    config = {
        # 'url_resolver': lambda url: github_doc_root + url,
        'auto_toc_tree_section': 'Contents',
        'enable_eval_rst': True,
    }
    app.add_config_value('recommonmark_config', config, True)
    app.add_transform(AutoStructify)

    # from m2r to make `mdinclude` work
    app.add_config_value('no_underscore_emphasis', False, 'env')
    app.add_config_value('m2r_parse_relative_links', False, 'env')
    app.add_config_value('m2r_anonymous_references', False, 'env')
    app.add_config_value('m2r_disable_inline_math', False, 'env')
    app.add_directive('mdinclude', MdInclude)
    app.add_css_file('custom.css')

    # generate the components.rst file dynamically so it points to stable/latest
    # of subprojects correctly
    loader = FileSystemLoader(Path(__file__).resolve().parent / templates_path[0])
    env = Environment(loader=loader)
    template = env.get_template('components.rst')
    with open(Path(__file__).resolve().parent / 'components.rst', 'w') as f:
        f.write(
            template.render(
                intersphinx_mapping=intersphinx_mapping,
                programs=labscript_suite_programs,
                current_project=project,
                img_path=img_path
            )
        )

    # hooks to test docstring coverage
    app.connect('autodoc-process-docstring', doc_coverage)
    app.connect('build-finished', doc_report)


members_to_watch = ['module', 'class', 'function', 'exception', 'method', 'attribute']
doc_count = 0
undoc_count = 0
undoc_objects = []
undoc_print_objects = False


def doc_coverage(app, what, name, obj, options, lines):
    global doc_count
    global undoc_count
    global undoc_objects

    if (what in members_to_watch and len(lines) == 0):
        # blank docstring detected
        undoc_count += 1
        undoc_objects.append(name)
    else:
        doc_count += 1


def doc_report(app, exception):
    global doc_count
    global undoc_count
    global undoc_objects
    # print out report of documentation coverage
    total_docs = undoc_count + doc_count
    if total_docs != 0:
        print(f'\nAPI Doc coverage of {doc_count/total_docs:.1%}')
        if undoc_print_objects or os.environ.get('READTHEDOCS'):
            print('\nItems lacking documentation')
            print('===========================')
            print(*undoc_objects, sep='\n')
    else:
        print('No docs counted, run \'make clean\' then rebuild to get the count.')


#####################################################################
#                                                                   #
# /plugins/progress_bar/__init__.py                                 #
#                                                                   #
# Copyright 2018, Christopher Billington                            #
#                                                                   #
# This file is part of the program BLACS, in the labscript suite    #
# (see http://labscriptsuite.org), and is licensed under the        #
# Simplified BSD License. See the license.txt file in the root of   #
# the project for the full license.                                 #
#                                                                   #
#####################################################################
import logging
import os
import subprocess
import threading
import sys
import time
from queue import Queue, Empty

import numpy as np

from qtutils import UiLoader, inmain, inmain_decorator
from qtutils.qt import QtGui, QtWidgets, QtCore

import labscript_utils.h5_lock
import h5py

import labscript_utils.properties as properties
from labscript_utils.connections import ConnectionTable
from zprocess import TimeoutError
from labscript_utils.ls_zprocess import Event
from blacs.plugins import PLUGINS_DIR, callback

name = "Progress Bar"
module = "progress_bar" # should be folder name
logger = logging.getLogger('BLACS.plugin.%s'%module)

# The progress bar will update every UPDATE_INTERVAL seconds, or at the marker
# times, whichever is soonest after the last update:
UPDATE_INTERVAL = 0.02
BAR_MAX = 1000

def _ensure_str(s):
    """convert bytestrings and numpy strings to python strings"""
    return s.decode() if isinstance(s, bytes) else str(s)


def black_has_good_contrast(r, g, b):
    """Return whether black text or white text would have better contrast on a
    background of the given colour, according to W3C recommendations (see
    https://www.w3.org/TR/WCAG20/). Return True for black or False for
    white"""
    cs = []
    for c in r, g, b:
        c = c / 255.0
        if c <= 0.03928:
            c = c/12.92
        else:
            c = ((c+0.055)/1.055) ** 2.4
        cs.append(c)
    r, g, b = cs
    L = 0.2126 * r + 0.7152 * g + 0.0722 * b
    return L > np.sqrt(1.05 * 0.05) - 0.05


class Plugin(object):
    def __init__(self, initial_settings):
        self.menu = None
        self.notifications = {}
        self.initial_settings = initial_settings
        self.BLACS = None
        self.command_queue = Queue()
        self.master_pseudoclock = None
        self.shot_start_time = None
        self.stop_time = None
        self.markers = None
        self.waits = None
        self.time_spent_waiting = None
        self.next_wait_index = None
        self.next_marker_index = None
        self.bar_text_prefix = None
        self.h5_filepath = None
        self.wait_completed_events_supported = False
        self.wait_completed = Event('wait_completed', role='wait')
        self.mainloop_thread = threading.Thread(target=self.mainloop)
        self.mainloop_thread.daemon = True
        
    def plugin_setup_complete(self, BLACS):
        self.BLACS = BLACS
        self.ui = UiLoader().load(os.path.join(PLUGINS_DIR, module, 'controls.ui'))
        self.bar = self.ui.bar
        self.style = QtWidgets.QStyleFactory.create('Fusion')
        if self.style is None:
            # If we're on Qt4, fall back to Plastique style:
            self.style = QtWidgets.QStyleFactory.create('Plastique')
        if self.style is None:
            # Not sure what's up, but fall back to app's default style:
            self.style = QtWidgets.QApplication.style()
        self.bar.setStyle(self.style)
        self.bar.setMaximum(BAR_MAX)
        self.bar.setAlignment(QtCore.Qt.AlignCenter)
        # Add our controls to the BLACS gui:
        BLACS['ui'].queue_status_verticalLayout.insertWidget(0, self.ui)
        # We need to know the name of the master pseudoclock so we can look up
        # the duration of each shot:
        self.master_pseudoclock = self.BLACS['experiment_queue'].master_pseudoclock

        # Check if the wait monitor device, if any, supports wait completed events:
        with h5py.File(self.BLACS['connection_table_h5file'], 'r') as f:
            if 'waits' in f:
                acq_device = f['waits'].attrs['wait_monitor_acquisition_device']
                acq_device = _ensure_str(acq_device)
                if acq_device:
                    props = properties.get(f, acq_device, 'connection_table_properties')
                    if props.get('wait_monitor_supports_wait_completed_events', False):
                        self.wait_completed_events_supported = True

        self.ui.wait_warning.hide()
        self.mainloop_thread.start()

    def get_save_data(self):
        return {}
    
    def get_callbacks(self):
        return {'science_over': self.on_science_over,
                'science_starting': self.on_science_starting}
        
    # This callback should be run after any callbacks that might have significant time
    # delays. For example it should run after the cycle_time science_starting callback.
    # The priority should be set accordingly.
    @callback(priority=200)
    def on_science_starting(self, h5_filepath):
        # Tell the mainloop that we're starting a shot:
        self.command_queue.put(('start', h5_filepath))

    @callback(priority=5)
    def on_science_over(self, h5_filepath):
        # Tell the mainloop we're done with this shot:
        self.command_queue.put(('stop', None))

    @inmain_decorator(True)
    def clear_bar(self):
        self.bar.setEnabled(False)
        self.bar.setFormat('No shot running')
        self.bar.setValue(0)
        self.bar.setPalette(self.style.standardPalette())
        self.ui.wait_warning.hide()

    def get_next_thing(self):
        """Figure out what's going to happen next: a wait, a time marker, or a
        regular update. Return a string saying which, and a float saying how
        long from now it will occur. If the thing has already happened but not
        been taken into account by our processing yet, then return zero for
        the time."""
        if self.waits is not None and self.next_wait_index < len(self.waits):
            next_wait_time = self.waits['time'][self.next_wait_index]
        else:
            next_wait_time = np.inf
        if self.markers is not None and self.next_marker_index < len(self.markers):
            next_marker_time = self.markers['time'][self.next_marker_index]
        else:
            next_marker_time = np.inf
        assert self.shot_start_time is not None
        assert self.time_spent_waiting is not None
        labscript_time = time.time() - self.shot_start_time - self.time_spent_waiting
        next_update_time = labscript_time + UPDATE_INTERVAL
        if next_update_time < next_wait_time and next_update_time < next_marker_time:
            return 'update', UPDATE_INTERVAL
        elif next_wait_time < next_marker_time:
            return 'wait', max(0, next_wait_time - labscript_time)
        else:
            return 'marker', max(0, next_marker_time - labscript_time)

    @inmain_decorator(True)
    def update_bar_style(self, marker=False, wait=False, previous=False):
        """Update the bar's style to reflect the next marker or wait,
        according to self.next_marker_index or self.next_wait_index. If
        previous=True, instead update to reflect the current marker or
        wait."""
        assert not (marker and wait)
        # Ignore requests to reflect markers or waits if there are no markers
        # or waits in this shot:
        marker = marker and self.markers is not None and len(self.markers) > 0
        wait = wait and self.waits is not None and len(self.waits) > 0
        if marker:
            marker_index = self.next_marker_index
            if previous:
                marker_index -= 1
                assert marker_index >= 0
            label, _, color = self.markers[marker_index]
            self.bar_text_prefix = '[%s] ' % _ensure_str(label)
            r, g, b = color[0]
            if color.dtype == np.uint8 and (r, g, b) == (0, 0, 0):
                # is old colour spec, in which (0,0,0) meant no colour specified.
                r, g, b = (-1, -1, -1)
            # (-1,-1,-1) means no colour set for the marker. Don't change the bar colour
            # in this case.
            if (r, g, b) != (-1, -1, -1):
                bar_color = QtGui.QColor(r, g, b)
                if black_has_good_contrast(r, g, b):
                    highlight_text_color = QtCore.Qt.black
                else:
                    highlight_text_color = QtCore.Qt.white
            else:
                bar_color = None
                highlight_text_color = None
            regular_text_color = None # use default
        elif wait:
            wait_index = self.next_wait_index
            if previous:
                wait_index -= 1
                assert wait_index >= 0
            label = self.waits[wait_index]['label']
            self.bar_text_prefix = '-%s- ' % _ensure_str(label)
            highlight_text_color = regular_text_color = QtGui.QColor(192, 0, 0)
            bar_color = QtCore.Qt.gray
        if marker or wait:
            palette = self.style.standardPalette()
            if bar_color is not None:
                palette.setColor(QtGui.QPalette.Highlight, bar_color)
            # Ensure the colour of the text on the filled in bit of the progress
            # bar has good contrast:
            if highlight_text_color is not None:
                palette.setColor(QtGui.QPalette.HighlightedText, highlight_text_color)
            if regular_text_color is not None:
                palette.setColor(QtGui.QPalette.Text, regular_text_color)
            self.bar.setPalette(palette)
        else:
            self.bar_text_prefix = None
            # Default palette:
            self.bar.setPalette(self.style.standardPalette())

    @inmain_decorator(True)
    def update_bar_value(self, marker=False, wait=False):
        """Update the progress bar with the current time elapsed. If marker or wait is
        true, then use the exact time at which the next marker or wait is defined,
        rather than the current time as returned by time.time()"""
        thinspace = u'\u2009'
        self.bar.setEnabled(True)
        assert not (marker and wait)
        if marker:
            labscript_time = self.markers['time'][self.next_marker_index]
        elif wait:
            labscript_time = self.waits['time'][self.next_wait_index]
        else:
            labscript_time = time.time() - self.shot_start_time - self.time_spent_waiting
        value = int(round(labscript_time / self.stop_time * BAR_MAX))
        self.bar.setValue(value)

        text = u'%.2f%ss / %.2f%ss (%%p%s%%)'
        text = text % (labscript_time, thinspace, self.stop_time, thinspace, thinspace)
        if self.bar_text_prefix is not None:
            text = self.bar_text_prefix + text
        self.bar.setFormat(text)

    def _start(self, h5_filepath):
        """Called from the mainloop when starting a shot"""
        self.h5_filepath = h5_filepath
        # Get the stop time, any waits and any markers from the shot:
        with h5py.File(h5_filepath, 'r') as f:
            props = properties.get(f, self.master_pseudoclock, 'device_properties')
            self.stop_time = props['stop_time']
            try:
                self.markers = f['time_markers'][:]
                self.markers.sort(order='time')
            except KeyError:
                self.markers = None
            try:
                self.waits = f['waits'][:]
                self.waits.sort(order='time')
            except KeyError:
                self.waits = None
        self.shot_start_time = time.time()
        self.time_spent_waiting = 0
        self.next_marker_index = 0
        self.next_wait_index = 0

    def _stop(self):
        """Called from the mainloop when ending a shot"""
        self.h5_filepath = None
        self.shot_start_time = None
        self.stop_time = None
        self.markers = None
        self.waits = None
        self.time_spent_waiting = None
        self.next_wait_index = None
        self.next_marker_index = None
        self.bar_text_prefix = None

    def mainloop(self):
        running = False
        self.clear_bar()
        while True:
            try:
                if running:
                    # How long until the next thing of interest occurs, and
                    # what is it? It can be either a wait, a marker, or a
                    # regular update.
                    next_thing, timeout = self.get_next_thing()
                    try:
                        command, _ = self.command_queue.get(timeout=timeout)
                    except Empty:
                        if next_thing == 'update':
                            self.update_bar_value()
                        if next_thing == 'marker':
                            self.update_bar_style(marker=True)
                            self.update_bar_value(marker=True)
                            self.next_marker_index += 1
                        elif next_thing == 'wait':
                            wait_start_time = time.time()
                            self.update_bar_style(wait=True)
                            self.update_bar_value(wait=True)
                            self.next_wait_index += 1
                            # wait for the wait to complete, but abandon
                            # processing if the command queue is non-empty,
                            # i.e. if a stop command is sent.
                            while self.command_queue.empty():
                                try:
                                    # Wait for only 0.1 sec at a time, so that
                                    # we can check if the queue is empty in between:
                                    self.wait_completed.wait(self.h5_filepath, timeout=0.1)
                                except TimeoutError:
                                    # Only wait for wait completed events if the wait
                                    # monitor device supports them. Otherwise, skip
                                    # after this first timeout, and it will just look
                                    # like the wait had 0.1 sec duration.
                                    if self.wait_completed_events_supported:
                                        # The wait is still in progress:
                                        continue
                                # The wait completed (or completion events are not
                                # supported):
                                self.time_spent_waiting += time.time() - wait_start_time
                                # Set the bar style back to whatever the
                                # previous marker was, if any:
                                self.update_bar_style(marker=True, previous=True)
                                self.update_bar_value()
                                break
                        continue
                else:
                    command, h5_filepath = self.command_queue.get()
                if command == 'close':
                    break
                elif command == 'start':
                    assert not running
                    running = True
                    self._start(h5_filepath)
                    self.update_bar_value()
                    if (
                        self.waits is not None
                        and len(self.waits) > 0
                        and not self.wait_completed_events_supported
                    ):
                        inmain(self.ui.wait_warning.show)
                elif command == 'stop':
                    self.clear_bar()
                    running = False
                    self._stop()
                else:
                    raise ValueError(command)
            except Exception:
                logger.exception("Exception in mainloop, ignoring.")
                # Stop processing of the current shot, if any.
                self.clear_bar()
                inmain(self.bar.setFormat, "Error in progress bar plugin")
                running = False
                self._stop()
    
    def close(self):
        self.command_queue.put(('close', None))
        self.mainloop_thread.join()

    # The rest of these are boilerplate:
    def get_menu_class(self):
        return None
        
    def get_notification_classes(self):
        return []
        
    def get_setting_classes(self):
        return []
    
    def set_menu_instance(self, menu):
        self.menu = menu
        
    def set_notification_instances(self, notifications):
        self.notifications = notifications
        
    



#####################################################################
#                                                                   #
# /compile_and_restart.py                                           #
#                                                                   #
# Copyright 2013, Monash University                                 #
#                                                                   #
# This file is part of the program BLACS, in the labscript suite    #
# (see http://labscriptsuite.org), and is licensed under the        #
# Simplified BSD License. See the license.txt file in the root of   #
# the project for the full license.                                 #
#                                                                   #
#####################################################################
import os
import shutil

from qtutils.qt.QtCore import *
from qtutils.qt.QtGui import *
from qtutils.qt.QtWidgets import *

from qtutils import *
import runmanager
from labscript_utils.qtwidgets.outputbox import OutputBox

from blacs import BLACS_DIR


class CompileAndRestart(QDialog):
    def __init__(self, blacs, globals_files,connection_table_labscript, output_path, close_notification_func=None):
        QDialog.__init__(self,blacs['ui'])
        self.setAttribute(Qt.WA_DeleteOnClose, True) # make sure the dialog is deleted when the window is closed
        
        self.globals_files = globals_files
        self.labscript_file = connection_table_labscript
        self.output_path = output_path
        self.tempfilename = self.output_path.strip('.h5')+'.temp.h5'
        self.blacs = blacs
        self.close_notification_func = close_notification_func
        
        self.ui = UiLoader().load(os.path.join(BLACS_DIR, 'compile_and_restart.ui'))
        self.output_box = OutputBox(self.ui.verticalLayout)       
        self.ui.restart.setEnabled(False)
        
        # Connect buttons
        self.ui.restart.clicked.connect(self.restart)
        self.ui.compile.clicked.connect(self.compile)
        self.ui.cancel.clicked.connect(self.reject)
        
        self.setLayout(self.ui.layout())
        self.resize(500, 300)
        self.show()
        self.setWindowTitle('Recompile connection table')
        self.compile()

    def closeEvent(self,event):
        if not self.ui.cancel.isEnabled():        
            event.ignore()            
        else:
            event.accept()
    
    def on_activate_default(self,window):
        if self.button_restart.get_sensitive():
            self.restart()
        elif self.button_compile.get_sensitive():
            self.compile()
                
    def compile(self):
        self.ui.compile.setEnabled(False)
        self.ui.cancel.setEnabled(False)
        self.ui.restart.setEnabled(False)
        msg = 'Recompiling connection table'
        self.ui.label.setText(msg)
        self.output_box.output(msg + '\n')
        runmanager.compile_labscript_with_globals_files_async(self.labscript_file,
            self.globals_files, self.tempfilename, self.output_box.port, self.finished_compiling)
    
    @inmain_decorator(True)    
    def finished_compiling(self, success):
        self.ui.compile.setEnabled(True)
        self.ui.cancel.setEnabled(True)
        if success:
            try:
                shutil.move(self.tempfilename,self.output_path)
            except Exception as e:
                msg = ('Couldn\'t replace existing connection table h5 file. ' + 
                       'Is it open in another process? ' +
                       'error was:\n %s\n') % str(e)
                self.output_box.output(msg, red=True)
                self.ui.label.setText('Compilation failed.')
                self.ui.restart.setEnabled(False)
                os.remove(self.tempfilename)
            else:
                self.ui.restart.setEnabled(True)
                self.ui.cancel.setEnabled(False)
                msg = 'Compilation succeeded, restart when ready'
                self.ui.label.setText(msg)
                self.output_box.output(msg + '\n')
        else:
            self.ui.restart.setEnabled(False)
            msg = 'Compilation failed. Please fix the errors in the connection table (python file) and try again'
            self.ui.label.setText(msg)
            self.output_box.output(msg + '\n')
            try:
                os.remove(self.tempfilename)
            except Exception:
                pass

    def restart(self):
        if self.close_notification_func:
            self.close_notification_func()
        QTimer.singleShot(100, self.blacs['ui'].close)
        self.accept()
        self.blacs['set_relaunch'](True)



#####################################################################
#                                                                   #
# /hardware_interfaces/output_classes.py                            #
#                                                                   #
# Copyright 2013, Monash University                                 #
#                                                                   #
# This file is part of the program BLACS, in the labscript suite    #
# (see http://labscriptsuite.org), and is licensed under the        #
# Simplified BSD License. See the license.txt file in the root of   #
# the project for the full license.                                 #
#                                                                   #
#####################################################################
import logging
import math
import sys

from qtutils.qt.QtCore import *
from qtutils.qt.QtGui import *
from qtutils.qt.QtWidgets import *


from labscript_utils.qtwidgets.analogoutput import AnalogOutput
from labscript_utils.qtwidgets.digitaloutput import DigitalOutput, InvertedDigitalOutput
from labscript_utils.qtwidgets.ddsoutput import DDSOutput
from labscript_utils.qtwidgets.imageoutput import ImageOutput
from labscript_utils.unitconversions import get_unit_conversion_class


class AO(object):
    def __init__(self, hardware_name, connection_name, device_name, program_function, settings, calib_class, calib_params, default_units, min, max, step, decimals):
        self._connection_name = connection_name
        self._hardware_name = hardware_name
        self._device_name = device_name
        
        self._locked = False
        self._comboboxmodel = QStandardItemModel()
        self._widgets = []
        self._current_units = default_units
        self._base_unit = default_units
        self._program_device = program_function
        
        # All of these are in base units ALWAYS
        self._current_value = 0 # value in base units
        self._current_step_size = step # step size in current units
        self._step_size = step # step size in base units
        self._limits = [min,max]
        self._decimals = decimals
                
        self._logger = logging.getLogger('BLACS.%s.%s'%(self._device_name,hardware_name)) 
        
        # Initialise Calibrations
        self._comboboxmodel.appendRow(QStandardItem(self._base_unit))
        if calib_class is not None:
            try:
                cls = get_unit_conversion_class(calib_class)
            # ImportError if module doesn't exist. Attribute error if it does but the
            # class name does not exist within it. KeyError if the unit conversion class
            # is an old-style unqualified class name expected to be in the globals()
            # dict of the unitconversions module, but does not exist.
            except (ImportError, AttributeError, KeyError):
                cls = None
            if cls is None or not isinstance(calib_params, dict) or cls.base_unit != default_units:
                # log an error:  
                reason = ''
                if cls is None:
                    reason = f'The unit conversion class {calib_class} could not be imported. Ensure it is available on the computer running BLACS.'
                elif not isinstance(calib_params, dict):
                    reason = 'The parameters for the unit conversion class are not a dictionary. Check your connection table code for errors and recompile it'
                elif cls.base_unit != default_units:
                    reason = f'The base unit of your unit conversion class does not match this hardware channel. The hardware channel has base units {default_units} while your unit conversion class uses {cls.base_unit}'
                self._logger.error('The unit conversion class (%s) could not be loaded. Reason: %s'%(calib_class,reason))   
                # Use default units
                self._calibration = None
            else:
                try:
                    # initialise calibration class
                    self._calibration = cls(calib_params)  
                    self._logger.debug('unit conversion class instantiated')                    
                    for unit in self._calibration.derived_units:
                        try:
                            self._comboboxmodel.appendRow(QStandardItem(unit))
                        except Exception:
                             self._logger.exception('Error while trying to add unit "%s"'%unit)                    
                except Exception:
                    self._logger.exception('Error while trying to instantiate unit conversion class')
                    self._calibration = None
        else:
            # use default units
            self._calibration = None
            self._logger.debug('No unit conversion class specified')
        
        self._update_from_settings(settings,program=False)
    
    def _update_from_settings(self,settings,program=True):
        # Build up the settings dictionary if it isn't already
        if not isinstance(settings,dict):
            settings = {}
        if 'front_panel_settings' not in settings or not isinstance(settings['front_panel_settings'],dict):
            settings['front_panel_settings'] = {}
        if self._hardware_name not in settings['front_panel_settings'] or not isinstance(settings['front_panel_settings'][self._hardware_name],dict):
            settings['front_panel_settings'][self._hardware_name] = {}
        # Set default values if they are not already saved in the settings dictionary
        if 'base_value' not in settings['front_panel_settings'][self._hardware_name]:
            settings['front_panel_settings'][self._hardware_name]['base_value'] = False
        if 'locked' not in settings['front_panel_settings'][self._hardware_name]:
            settings['front_panel_settings'][self._hardware_name]['locked'] = False
        if 'base_step_size' not in settings['front_panel_settings'][self._hardware_name]:
            settings['front_panel_settings'][self._hardware_name]['base_step_size'] = self._step_size
        if 'current_units' not in settings['front_panel_settings'][self._hardware_name]:
            settings['front_panel_settings'][self._hardware_name]['current_units'] = self._base_unit
        if 'name' not in settings['front_panel_settings'][self._hardware_name]:
            settings['front_panel_settings'][self._hardware_name]['name'] = self._connection_name
        
    
        # only keep a reference to the part of the settings dictionary relevant to this DO
        self._settings = settings['front_panel_settings'][self._hardware_name]
    
        # Update the state of the button
        self.set_value(self._settings['base_value'],program=program)

        # Update the lock state
        self._update_lock(self._settings['locked'])
        
        # Update the step size
        self.set_step_size(self._settings['base_step_size'],self._base_unit)
    
        # Update the unit selection
        if self._calibration and self._settings['current_units'] in self._calibration.derived_units:
            self.change_unit(self._settings['current_units'],program=program)
        else:
            self.change_unit(self._base_unit,program=program)
     
    def convert_value_to_base(self, value, unit):
        if unit != self._base_unit:
            if self._calibration and unit in self._calibration.derived_units:
                return getattr(self._calibration,unit+"_to_base")(value)
             
            # TODO: include device name somehow, and also the calibration class name
            raise RuntimeError('The value %s (%s) could not be converted to base units because the hardware channel %s, named %s, either does not have a unit conversion class or the unit specified was invalid'%(str(value),unit,self._hardware_name,self._connection_name))
        else:
            return value
            
    def convert_value_from_base(self, value, unit):  
        if unit != self._base_unit:
            if self._calibration and unit in self._calibration.derived_units:
                return getattr(self._calibration,unit+"_from_base")(value)
        
            # TODO: include device name somehow, and also the calibration class name
            raise RuntimeError('The value %s (%s) could not be converted to base units because the hardware channel %s, named %s, either does not have a unit conversion class or the unit specified was invalid'%(str(value),unit,self._hardware_name,self._connection_name))
        else:
            return value
            
    # handles the conversion of a range centered on value in units to base units
    # In other words, how big is "range" in base units assuming that you care about what range is
    # between value-range/2 and value+range/2
    #
    # If value+range/2 or value-range/2 is outside of the limits, then we will shift the fraction of range
    # used on the offending side of value
    # If range is greater than the difference of the limits, we will return the difference between the limits
    def convert_range_to_base(self,value,range,unit):
        self._logger.debug('convert_range_to_base called. value: %f, range: %f, unit: %s'%(value,range,unit))
        
        # Do we need to convert the limits?
        if unit != self._base_unit:
            limits = [self.convert_value_from_base(self._limits[0],unit),self.convert_value_from_base(self._limits[1],unit)]
            if limits[0] > limits[1]:
                limits[0],limits[1] = limits[1],limits[0]
        else:
            limits = self._limits
        
        self._logger.debug('limits in unit: %s), limits=[%f,%f]'%(unit,limits[0],limits[1]))
        
        # limits are now in the units given to the function!
        # (As are range and value)
            
        # If range is bigger than the difference of the limits, return the difference of the limits
        # in base units
        if range >= abs(limits[0]-limits[1]):
            limits = [self.convert_value_to_base(limits[0],unit),self.convert_value_to_base(limits[1],unit)]            
            self._logger.debug('range bigger than range of limits, returning difference of limits') 
            return abs(limits[0]-limits[1])
          
        # At this point, the range must fit inside the limits, so if we find we are out of bounds on one side, 
        # we can be certain shifting the fractions will not cause us to go out of bounds on the other side
        positive_fraction = range/2.0
        negative_fraction = range/2.0
        self._logger.debug('fractions are.... positive_fraction: %f, negative_fraction: %f'%(positive_fraction, negative_fraction))
        
        # If the value+range/2 is greater than the upper limit, shift the fraction 
        if value+positive_fraction > limits[1]:           
            positive_fraction = abs(limits[1]-value)
            negative_fraction = abs(range-positive_fraction) 
            self._logger.debug('outside upper limit. positive_fraction: %f, negative_fraction: %f'%(positive_fraction, negative_fraction))
        # Similarly if value-range/2 is less than the lower limit, shift the fraction
        elif value-negative_fraction < limits[0]:    
            negative_fraction = abs(limits[0]-value)        
            positive_fraction = abs(range-negative_fraction)
            self._logger.debug('outside lower limit. positive_fraction: %f, negative_fraction: %f'%(positive_fraction, negative_fraction))
        
        self._logger.debug('converting values to base units')            
        # Now do the conversion!
        bound1 = self.convert_value_to_base(value+positive_fraction,unit)
        bound2 = self.convert_value_to_base(value-negative_fraction,unit)
        self._logger.debug('range in base units is: %f'%(abs(bound1-bound2)))
        
        return abs(bound1-bound2)
    
    # This does the reverse of teh above function, with the same rules
    def convert_range_from_base(self,value,range,unit):
        # limits are always in base units
        limits = self._limits
        
        # limits are now in base units!
        # (As are range and value)
            
        # If range is bigger than the difference of the limits, return the difference of the limits
        # in the specified units units
        if range >= abs(limits[0]-limits[1]):
            limits = [self.convert_value_from_base(limits[0],unit),self.convert_value_from_base(limits[1],unit)]
            return abs(limits[0]-limits[1])
          
        # At this point, the range must fit inside the limits, so if we find we are out of bounds on one side, 
        # we can be certain shifting the fractions will not cause us to go out of bounds on the other side
        positive_fraction = range/2.0
        negative_fraction = range/2.0
        # If the value+range/2 is greater than the upper limit, shift the fraction 
        if value+positive_fraction > limits[1]:
            positive_fraction = abs(limits[1]-value)
            negative_fraction = abs(range-positive_fraction)
        # Similarly if value-range/2 is less than the lower limit, shift the fraction
        elif value-negative_fraction < limits[0]:    
            negative_fraction = abs(limits[0]-value)        
            positive_fraction = abs(range-negative_fraction)
        
        # Now do the conversion!
        bound1 = self.convert_value_from_base(value+positive_fraction,unit)
        bound2 = self.convert_value_from_base(value-negative_fraction,unit)
        
        return abs(bound1-bound2)

    def create_widget(self,display_name=None, horizontal_alignment=False, parent=None):
        widget = AnalogOutput(self._hardware_name,self._connection_name,display_name, horizontal_alignment, parent)
        self.add_widget(widget)
        return widget
        
    def add_widget(self, widget):
        if widget in self._widgets:
            return False
            
        self._widgets.append(widget)
    
        # make sure the widget knows about this AO. 
        widget.set_AO(self,notify_old_AO=True,notify_new_AO=False)
        
        # Now connect this widgets signal to the AO slot
        widget.connect_value_change(self.set_value)
        
        # set the properties of the widgets...
        # set comboboxmodel
        widget.block_combobox_signals()
        widget.set_combobox_model(self._comboboxmodel)
        widget.unblock_combobox_signals()
        # This will set the min/max/value/num.decimals/stepsie and current_unit of ALL widgets
        # including the one just added!
        self.change_unit(self._current_units,program=False) # don't need to program though!
        # This will update the lock state of ALL widgets, including the one just added!
        self._update_lock(self._locked)
        
        return True
    
    # If calling this method directly from outside the set_AO function in the analog widget
    # you should NOT specify a value for new_AO.
    def remove_widget(self,widget,call_set_AO = True,new_AO = None):
        if widget not in self._widgets:
            raise RuntimeError('The widget cannot be removed because it is not registered with this AO object')
            #TODO: Make the above error message better!
         
        self._widgets.remove(widget)  
        
        if call_set_AO:
            widget.set_AO(new_AO,True,True)
            
        # Further cleanup
        widget.disconnect_value_change()
        widget.set_combobox_model(QStandardItemModel())
        
    def change_unit(self,unit,program=True):     
        # These values are always stored in base units!
        property_value_list = [self._current_value,self._limits[0],self._limits[1]]
        property_range_list = [self._step_size]
        
        self._logger.debug('changing unit to %s'%unit)
        self._logger.debug('Values in base units are: value: %f, lower_limit: %f, upper_limit: %f'%(property_value_list[0],property_value_list[1],property_value_list[2]))
        self._logger.debug('ranges in base units are: step_size: %f'%(property_range_list[0]))
        
        # Now convert to the new unit
        if unit != self._base_unit:
            for index,param in enumerate(property_value_list):
                #convert each to base units
                property_value_list[index] = self.convert_value_from_base(param,unit)
            for index,param in enumerate(property_range_list):
                #convert each to base units
                property_range_list[index] = self.convert_range_from_base(self._current_value,param,unit)
            
            self._logger.debug('Values in new unit are: value: %f, lower_limit: %f, upper_limit: %f'%(property_value_list[0],property_value_list[1],property_value_list[2]))
            self._logger.debug('ranges in new unit are: step_size: %f'%(property_range_list[0]))        
            
            # figure out how many decimal points we need in the new unit
            smallest_step = 10**(-self._decimals)
            self._logger.debug('Smallest step size in base units: %f'%smallest_step)
            smallest_step_in_new_unit = self.convert_range_from_base(self._current_value+smallest_step,smallest_step,unit)
            self._logger.debug('Smallest step size in new_unit: %f'%smallest_step_in_new_unit)
            try:
                if smallest_step_in_new_unit > 1:
                    if smallest_step_in_new_unit > 10:
                        num_decimals = 0
                    else:
                        num_decimals = 1
                else:
                    num_decimals = abs(math.floor(math.log10(smallest_step_in_new_unit))-2)
            except Exception:
                self._logger.warning('Failed to convert number of significant figures to new unit. Loss of precision likely (in manual mode) for this unit. Probably cause is a unit conversion class that imposes limits on the converted values.')
                num_decimals = self._decimals
        else:
            num_decimals = self._decimals
        
        # Store the current units
        self._current_units = unit  
        self._settings['current_units'] = unit    
        
        # Check to see if the upper/lower bound has switched
        if property_value_list[1] > property_value_list[2]:
            property_value_list[1], property_value_list[2] = property_value_list[2], property_value_list[1]
        
        # Now update all the widgets
        for widget in self._widgets:
            # Update the combo box
            widget.block_combobox_signals()
            widget.set_selected_unit(unit)
            widget.unblock_combobox_signals()
            
            # block the spinbox from emitting a signal
            widget.block_spinbox_signals()
            # Update the limits
            widget.set_limits(property_value_list[1],property_value_list[2])
            # Update the step size
            widget.set_step_size(property_range_list[0])
            # Update the decimals
            widget.set_num_decimals(num_decimals)
            # Update the value - This should be the last thing you do, 
            #                    otherwise it might get truncated or 
            #                    limited in a bad way
            widget.set_spinbox_value(property_value_list[0],unit)
            # unblock the spinbox signals
            widget.unblock_spinbox_signals()
      
    @property
    def value(self):
        return self._current_value
        
    def set_value(self, value, unit=None, program=True):
        # conversion to float means a string can be passed in too:
        value = float(value)
        
        if unit is not None and unit != self._base_unit:
            self._current_value = self.convert_value_to_base(value,unit)
        else:
            self._current_value = value
        
        # Update the saved value in the settings dictionary
        self._settings['base_value'] = self._current_value
            
        if program:
            self._logger.debug('program device called')
            self._program_device()
            
        for widget in self._widgets:
            # block signals
            widget.block_spinbox_signals()
            # update widget
            widget.set_spinbox_value(value,unit if unit is not None else self._base_unit)
            # unblock signals            
            widget.unblock_spinbox_signals()
    
    def set_step_size(self,step_size,unit):
        self._logger.debug('set_step_size called. step_size: %f, unit: %s'%(step_size,unit))
        if unit != self._base_unit:
            # convert and store!
            value = self.convert_value_from_base(self._current_value,unit)
            self._step_size = self.convert_range_to_base(value,step_size,unit)
        else:
            # This check is usually performed when converting the range to base units
            # But since we are already in base units we should do it here
            if abs(self._limits[0]-self._limits[1]) <= step_size:
                step_size = abs(self._limits[0]-self._limits[1])
            self._step_size = step_size
        
        self._logger.debug('step_size in base units: %f'%self._step_size)
        
        #self._current_step_size = self._step_size
        self._settings['base_step_size'] = self._step_size
        
        # now convert to current units
        self._current_step_size = self.get_step_size(self._current_units)        
        self._logger.debug('step_size in current units (%s): %f'%(self._current_units,self._current_step_size))
    
        # Update the step size for all widgets
        for widget in self._widgets:
            widget.set_step_size(self._current_step_size)
    
    def get_step_size(self,unit):
        if unit != self._base_unit:
            # we should convert it
            return self.convert_range_from_base(self._current_value,self._step_size,unit)
        else:
            return self._step_size
    
    def lock(self):
        self._update_lock(True)
        
    def unlock(self):
        self._update_lock(False)
    
    def _update_lock(self, locked):    
        self._locked = locked        
        self._settings['locked'] = locked
        
        # Lock all widgets if they are not already locked
        for widget in self._widgets:
            if locked:
                widget.lock(False)
            else:
                widget.unlock(False)

    @property
    def name(self):
        return self._hardware_name + ' - ' + self._connection_name
            
class DO(object):
    def __init__(self, hardware_name, connection_name, device_name, program_function, settings):
        self._hardware_name = hardware_name
        self._connection_name = connection_name
        self._widget_list = []
        
        self._device_name = device_name
        self._logger = logging.getLogger('BLACS.%s.%s'%(self._device_name,hardware_name))

        # Note that while we could store self._current_state and self._locked in the
        # settings dictionary, this dictionary is available to other parts of BLACS
        # and using separate variables avoids those parts from being able to directly
        # influence behaviour (the worst they can do is change the value used on initialisation)
        self._locked = False
        self._current_state = False
        self._program_device = program_function
        self._update_from_settings(settings)
    
    def _update_from_settings(self,settings):
        # Build up the settings dictionary if it isn't already
        if not isinstance(settings,dict):
            settings = {}
        if 'front_panel_settings' not in settings or not isinstance(settings['front_panel_settings'],dict):
            settings['front_panel_settings'] = {}
        if self._hardware_name not in settings['front_panel_settings'] or not isinstance(settings['front_panel_settings'][self._hardware_name],dict):
            settings['front_panel_settings'][self._hardware_name] = {}
        # Set default values if they are not already saved in the settings dictionary
        if 'base_value' not in settings['front_panel_settings'][self._hardware_name]:
            settings['front_panel_settings'][self._hardware_name]['base_value'] = False
        if 'locked' not in settings['front_panel_settings'][self._hardware_name]:
            settings['front_panel_settings'][self._hardware_name]['locked'] = False
        if 'name' not in settings['front_panel_settings'][self._hardware_name]:
            settings['front_panel_settings'][self._hardware_name]['name'] = self._connection_name
        
    
        # only keep a reference to the part of the settings dictionary relevant to this DO
        self._settings = settings['front_panel_settings'][self._hardware_name]
    
        # Update the state of the button
        self.set_value(self._settings['base_value'],program=False)

        # Update the lock state
        self._update_lock(self._settings['locked'])

    def create_widget(self, *args, **kwargs):
        inverted = kwargs.pop("inverted", False)
        if not inverted:
            widget = DigitalOutput('%s\n%s'%(self._hardware_name,self._connection_name),*args,**kwargs)
        else:
            widget = InvertedDigitalOutput('%s\n%s'%(self._hardware_name,self._connection_name),*args,**kwargs)
        self.add_widget(widget, inverted=inverted)
        return widget

    def add_widget(self, widget, inverted=False):
        if widget not in self._widget_list:
            widget.set_DO(self,True,False)
            widget.toggled.connect(self.set_value if not inverted else lambda state: self.set_value(not state))
            self._widget_list.append(widget)
            self.set_value(self._current_state,False)
            self._update_lock(self._locked)
            return True
        return False
        
    def remove_widget(self,widget):
        if widget not in self._widget_list:
            # TODO: Make this error better!
            raise RuntimeError('The widget specified was not part of the DO object')
        widget.toggled.disconnect(self.set_value)
        self._widget_list.remove(widget)
        
    @property  
    def value(self):
        return bool(self._current_state)
    
    def lock(self):
        self._update_lock(True)
        
    def unlock(self):
        self._update_lock(False)
    
    def _update_lock(self,locked):
        self._locked = locked
        for widget in self._widget_list:
            if locked:
                widget.lock(False)
            else:
                widget.unlock(False)
        
        # update the settings dictionary if it exists, to maintain continuity on tab restarts
        self._settings['locked'] = locked
            
    def set_value(self,state,program=True):
        # conversion to integer, then bool means we can safely pass in
        # either a string '1' or '0', True or False or 1 or 0
        state = bool(int(state))

        # We are programatically setting the state, so break the check lock function logic
        self._current_state = state
        
        # update the settings dictionary if it exists, to maintain continuity on tab restarts
        self._settings['base_value'] = state
        
        if program:            
            self._logger.debug('program device called')
            self._program_device()
            
        for widget in self._widget_list:
            if state != widget.state:
                widget.blockSignals(True)
                widget.state = state
                widget.blockSignals(False)
   
    @property
    def name(self):
        return self._hardware_name + ' - ' + self._connection_name
   

class Image(object):
    def __init__(self, hardware_name, connection_name, device_name, program_function, settings, width, height, x = None, y = None):
        self._hardware_name = hardware_name
        self._connection_name = connection_name
        self._widget_list = []
        
        if x is not None and y is not None:
            self.location = 'Location: (%d, %d) '%(x,y)
        else:
            self.location = ""
        
        self.width = width
        self.height = height
        
        
        self._device_name = device_name
        self._logger = logging.getLogger('BLACS.%s.%s'%(self._device_name,hardware_name)) 
                
        # Note that while we could store self._current_state and self._locked in the
        # settings dictionary, this dictionary is available to other parts of BLACS
        # and using separate variables avoids those parts from being able to directly
        # influence behaviour (the worst they can do is change the value used on initialisation)
        self._locked = False
        self._current_value = ""
        self._program_device = program_function
        self._update_from_settings(settings)
        
    def _update_from_settings(self, settings):
        # Build up the settings dictionary if it isn't already
        if not isinstance(settings,dict):
            settings = {}
        if 'front_panel_settings' not in settings or not isinstance(settings['front_panel_settings'],dict):
            settings['front_panel_settings'] = {}
        if self._hardware_name not in settings['front_panel_settings'] or not isinstance(settings['front_panel_settings'][self._hardware_name],dict):
            settings['front_panel_settings'][self._hardware_name] = {}
        # Set default values if they are not already saved in the settings dictionary
        if 'base_value' not in settings['front_panel_settings'][self._hardware_name]:
            settings['front_panel_settings'][self._hardware_name]['base_value'] = ""
        if 'locked' not in settings['front_panel_settings'][self._hardware_name]:
            settings['front_panel_settings'][self._hardware_name]['locked'] = False
        if 'name' not in settings['front_panel_settings'][self._hardware_name]:
            settings['front_panel_settings'][self._hardware_name]['name'] = self._connection_name
        
    
        # only keep a reference to the part of the settings dictionary relevant to this DO
        self._settings = settings['front_panel_settings'][self._hardware_name]
    
        # Update the state of the button
        self.set_value(self._settings['base_value'],program=False)

        # Update the lock state
        self._update_lock(self._settings['locked'])
        
    def create_widget(self, *args, **kwargs):
        widget = ImageOutput('%s (%s)\n%sDimensions(%d, %d)'%(self._hardware_name,self._connection_name, self.location, self.width, self.height),self.width, self.height, *args,**kwargs)
        self.add_widget(widget)
        return widget
        
    def add_widget(self, widget):
        if widget not in self._widget_list:
            widget.set_Image(self,True,False)
            widget.imageUpdated.connect(self.set_value)
            self._widget_list.append(widget)
            self.set_value(self._current_value,False)
            self._update_lock(self._locked)
            return True
        return False
        
    def remove_widget(self, widget):
        if widget not in self._widget_list:
            # TODO: Make this error better!
            raise RuntimeError('The widget specified was not part of the Image object')
        widget.imageUpdated.disconnect(self.set_value)
        self._widget_list.remove(widget)
        
    @property
    def value(self):
        return str(self._current_value)
        
    def lock(self):
        self._update_lock(True)
    
    def unlock(self):
        self._update_lock(False)
        
    def _update_lock(self, locked):
        self._locked = locked
        for widget in self._widget_list:
            if locked:
                widget.lock(False)
            else:
                widget.unlock(False)
        
        # update the settings dictionary if it exists, to maintain continuity on tab restarts
        self._settings['locked'] = locked
        
    def set_value(self, value, program = True):
        value = str(value)  
        
        # We are programatically setting the value, so break the check lock function logic
        self._current_value = value
        
        # update the settings dictionary if it exists, to maintain continuity on tab restarts
        self._settings['base_value'] = value
        
        if program:            
            self._logger.debug('program device called')
            self._program_device()
            
        for widget in self._widget_list:
            if value != widget.value:
                widget.blockSignals(True)
                widget.value = value
                widget.blockSignals(False)
        
    @property
    def name(self):
        return self._hardware_name + ' - ' + self._connection_name
   
class DDS(object):
    def __init__(self, hardware_name, connection_name, output_list):
        self._hardware_name = hardware_name
        self._connection_name = connection_name
        self._sub_channel_list = ['freq','amp','phase','gate']
        self._widget_list = []
        for subchnl in self._sub_channel_list:
            value = None
            if subchnl in output_list:
                value = output_list[subchnl]
            
                setattr(self,subchnl,value)
            
    def create_widget(self,*args,**kwargs):
        widget = DDSOutput(self._hardware_name,self._connection_name,*args,**kwargs)
        self.add_widget(widget)
        return widget
    
    def add_widget(self, widget):
        if widget in self._widget_list:
            return False
            
        # Check that the widget has a method for getting/showin/hiding subwidgets        
        for subchnl in self._sub_channel_list:
            widget.get_sub_widget(subchnl)
            widget.hide_sub_widget(subchnl)
            widget.show_sub_widget(subchnl)
            
        self._widget_list.append(widget)
        
        for subchnl in self._sub_channel_list:
            if hasattr(self,subchnl):
                getattr(self,subchnl).add_widget(widget.get_sub_widget(subchnl))            
                widget.show_sub_widget(subchnl)
            else:
                widget.hide_sub_widget(subchnl)
        
        return True
        
    def remove_widget(self,widget):
        if widget not in self._widget_list:
            # TODO: Make this error better!
            raise RuntimeError('The widget specified was not part of the DDS object')
        
        for subchnl in self._sub_channel_list:
            if hasattr(self,subchnl):
                getattr(self,subchnl).remove_widget(widget.get_sub_widget(subchnl))  
        
        self._widget_list.remove(widget)
    
    def get_subchnl_list(self):
        subchnls = []
        for subchnl in self._sub_channel_list:
            if hasattr(self,subchnl):
                subchnls.append(subchnl)
                
        return subchnls
        
    def get_unused_subchnl_list(self):
        return list(set(self._sub_channel_list).difference(set(self.get_subchnl_list())))
    
    @property
    def value(self):
        value = {}
        for subchnl in self._sub_channel_list:
            if hasattr(self,subchnl):
                value[subchnl] = getattr(self,subchnl).value
        return value
        
    def set_value(self,value,program=True):
        for subchnl in self._sub_channel_list:
            if subchnl in value:
                if hasattr(self,subchnl):
                    getattr(self,subchnl).set_value(value[subchnl],program=program)
                    
    @property
    def name(self):
        return self._hardware_name + ' - ' + self._connection_name
        
if __name__ == '__main__':
    from labscript_utils.qtwidgets.toolpalette import ToolPaletteGroup
    import sys
    
    qapplication = QApplication(sys.argv)
    
    window = QWidget()
    layout = QVBoxLayout(window)
    widget = QWidget()
    layout.addWidget(widget)
    tpg = ToolPaletteGroup(widget)
    toolpalette = tpg.append_new_palette('Digital Outputs')    
    toolpalette2 = tpg.append_new_palette('Analog Outputs')    
    layout.addItem(QSpacerItem(0,0,QSizePolicy.Minimum,QSizePolicy.MinimumExpanding))
    
    # create settings dictionary
    settings = {'front_panel_settings':{
                    'do0':{
                        'base_value':False,
                        'locked':False,
                        },
                    'ao0':{
                        'base_value':3.0,
                        'locked':False,
                        'base_step_size':0.1,
                        'current_units':'V',
                        }
                }
        }
    
    def print_something():
        print('program_function called')

    # Create a DO object
    my_DO = DO(hardware_name='do0', connection_name='my first digital output', program_function=print_something, settings=settings)
    
    # Link in two DO widgets
    button1 = DigitalOutput('do0\nmy first digital output')
    button2 = DigitalOutput('a linked do0')
    toolpalette.addWidget(button1)
    toolpalette.addWidget(button2)
    my_DO.add_widget(button1)
    my_DO.add_widget(button2)
    
    # Create an AO object
    my_AO = AO(hardware_name = 'ao0', connection_name='my ao', device_name='ni_blah',
                program_function=print_something, settings=settings, 
                calib_class=None, calib_params=None, default_units='V', 
                min=-10.0, max=10.0, step=0.01, decimals=3)
    
    # link in two AO widgets
    analog1 = AnalogOutput('AO1')
    analog2 = AnalogOutput('AO1 copy')
    my_AO.add_widget(analog1)
    my_AO.add_widget(analog2)
    toolpalette2.addWidget(analog1)
    toolpalette2.addWidget(analog2)
    
    # TODO: Add in test case for DDS
    
    
    window.show()
    sys.exit(qapplication.exec_())
    



#####################################################################
#                                                                   #
# /plugins/__init__.py                                              #
#                                                                   #
# Copyright 2013, Monash University                                 #
#                                                                   #
# This file is part of the program BLACS, in the labscript suite    #
# (see http://labscriptsuite.org), and is licensed under the        #
# Simplified BSD License. See the license.txt file in the root of   #
# the project for the full license.                                 #
#                                                                   #
#####################################################################
import os
import sys
import logging
import importlib
from types import MethodType
from collections import defaultdict
from labscript_utils.labconfig import LabConfig
from blacs import BLACS_DIR
PLUGINS_DIR = os.path.join(BLACS_DIR, 'plugins')

default_plugins = ['connection_table', 'general', 'theme']

logger = logging.getLogger('BLACS.plugins')

DEFAULT_PRIORITY = 10

class Callback(object):
    """Class wrapping a callable. At present only differs from a regular
    function in that it has a "priority" attribute - lower numbers means
    higher priority. If there are multiple callbacks triggered by the same
    event, they will be returned in order of priority by get_callbacks"""
    def __init__(self, func, priority=DEFAULT_PRIORITY):
        self.priority = priority
        self.func = func

    def __get__(self, instance, class_):
        """Make sure our callable binds like an instance method. Otherwise
        __call__ doesn't get the instance argument."""
        if instance is None:
            return self
        else:
            return MethodType(self, instance)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


class callback(object):
    """Decorator to turn a function into a Callback object. Presently
    optional, and only required if the callback needs to have a non-default
    priority set"""
    # Instantiate the decorator:
    def __init__(self, priority=DEFAULT_PRIORITY):
        self.priority = priority
    # Call the decorator
    def __call__(self, func):
        return Callback(func, self.priority)


def get_callbacks(name):
    """Return all the callbacks for a particular name, in order of priority"""
    import __main__
    BLACS = __main__.app
    callbacks = []
    for plugin in BLACS.plugins.values():
        try:
            plugin_callbacks = plugin.get_callbacks()
            if plugin_callbacks is not None:
                if name in plugin_callbacks:
                    callbacks.append(plugin_callbacks[name])
        except Exception:
            logger.exception('Error getting callbacks from %s.' % str(plugin))
            
    # Sort all callbacks by priority:
    callbacks.sort(key=lambda callback: getattr(callback, 'priority', DEFAULT_PRIORITY))
    return callbacks


exp_config = LabConfig()
if not exp_config.has_section('BLACS/plugins'):
    exp_config.add_section('BLACS/plugins')

modules = {}
for module_name in os.listdir(PLUGINS_DIR):
    if os.path.isdir(os.path.join(PLUGINS_DIR, module_name)) and module_name != '__pycache__':
        # is it a new plugin?
        # If so lets add it to the config
        if not module_name in [name for name, val in exp_config.items('BLACS/plugins')]:
            exp_config.set('BLACS/plugins', module_name, str(module_name in default_plugins))

        # only load activated plugins
        if exp_config.getboolean('BLACS/plugins', module_name):
            try:
                module = importlib.import_module('blacs.plugins.'+module_name)
            except Exception:
                logger.exception('Could not import plugin \'%s\'. Skipping.'%module_name)
            else:
                modules[module_name] = module



import os
from setuptools import setup

setup(
    use_scm_version={
        "version_scheme": "release-branch-semver",
        "local_scheme": os.getenv("SCM_LOCAL_SCHEME", "node-and-date"),
    }
)



#####################################################################
#                                                                   #
# /plugins/connection_table/__init__.py                             #
#                                                                   #
# Copyright 2013, Monash University                                 #
#                                                                   #
# This file is part of the program BLACS, in the labscript suite    #
# (see http://labscriptsuite.org), and is licensed under the        #
# Simplified BSD License. See the license.txt file in the root of   #
# the project for the full license.                                 #
#                                                                   #
#####################################################################
import logging
import os
import subprocess
import sys
import ast

from qtutils.qt.QtCore import *
from qtutils.qt.QtGui import *
from qtutils.qt.QtWidgets import *

from blacs.compile_and_restart import CompileAndRestart
from labscript_utils.filewatcher import FileWatcher
from qtutils import *
from blacs.plugins import PLUGINS_DIR

FILEPATH_COLUMN = 0
name = "Connection Table"
module = "connection_table" # should be folder name
logger = logging.getLogger('BLACS.plugin.%s'%module)

class Plugin(object):
    def __init__(self,initial_settings):
        self.menu = None
        self.notifications = {}
        self.initial_settings = initial_settings
        self.BLACS = None
        
    def get_menu_class(self):
        return Menu
        
    def get_notification_classes(self):
        return [RecompileNotification, BrokenDevicesNotification]

    def get_setting_classes(self):
        return [Setting]
        
    def get_callbacks(self):
        return {'settings_changed':self.notifications[RecompileNotification].setup_filewatching}

    def set_menu_instance(self,menu):
        self.menu = menu
        
    def set_notification_instances(self,notifications):
        self.notifications = notifications
        
    def plugin_setup_complete(self, BLACS):
        self.BLACS = BLACS
        # The 'clean' modified info. We don't save the 'dirty' modified info. If watched
        # files are 'dirty' then our callback will be immediately called.
        clean_modified_info = self.initial_settings.get('clean_modified_info', None)
        self.notifications[RecompileNotification].setup_filewatching(clean_modified_info)
        self.menu.close_notification_func = self.notifications[RecompileNotification].on_restart
        failed_devices = list(self.BLACS['experiment_queue'].BLACS.failed_device_settings.keys())
        if failed_devices:
            self.notifications[RecompileNotification]._show()
            self.notifications[BrokenDevicesNotification].set_broken_devices(failed_devices)
            self.notifications[BrokenDevicesNotification]._show()

    def get_save_data(self):
        return self.notifications[RecompileNotification].get_save_data()

    def close(self):
        self.notifications[RecompileNotification].close()
        self.notifications[BrokenDevicesNotification].close()

class Menu(object):
    def __init__(self,BLACS):
        self.BLACS = BLACS
        self.close_notification_func = None
        
    def get_menu_items(self):
        return {'name':name,        
                'menu_items':[{'name':'Edit',
                               'action':self.on_edit_connection_table,
                               'icon': ':/qtutils/fugue/document--pencil'
                              },
                              {'name':'Select Globals',
                               'action':self.on_select_globals,
                               'icon': ':/qtutils/fugue/table--pencil'
                              },
                              {'name':'Recompile',
                               'action':self.on_recompile_connection_table,
                               'icon': ':/qtutils/fugue/arrow-circle'
                              }
                             ]                                
               }
    
    def on_select_globals(self,*args,**kwargs):
        self.BLACS['settings'].create_dialog(goto_page=Setting)
      
    def on_edit_connection_table(self,*args,**kwargs):
        # get path to text editor
        editor_path = self.BLACS['exp_config'].get('programs','text_editor')
        editor_args = self.BLACS['exp_config'].get('programs','text_editor_arguments')
        if editor_path:  
            if '{file}' in editor_args:
                editor_args = editor_args.replace('{file}', self.BLACS['exp_config'].get('paths','connection_table_py'))
            else:
                editor_args = self.BLACS['exp_config'].get('paths','connection_table_py') + " " + editor_args            
            try:
                subprocess.Popen([editor_path,editor_args])
            except Exception:
                QMessageBox.information(self.BLACS['ui'],"Error","Unable to launch text editor. Check the path is valid in the experiment config file (%s) (you must restart BLACS if you edit this file)"%self.BLACS['exp_config'].config_path)
        else:
            QMessageBox.information(self.BLACS['ui'],"Error","No text editor path was specified in the experiment config file (%s) (you must restart BLACS if you edit this file)"%self.BLACS['exp_config'].config_path)
    
    def on_recompile_connection_table(self,*args,**kwargs):
        logger.info('recompile connection table called')
        # get list of globals
        globals_files = self.BLACS['settings'].get_value(Setting,'globals_list')
        # Remove unicode encoding so that zprocess.locking doesn't crash
        for i in range(len(globals_files)):
            globals_files[i] = str(globals_files[i])
        CompileAndRestart(self.BLACS, globals_files, self.BLACS['exp_config'].get('paths','connection_table_py'), self.BLACS['exp_config'].get('paths','connection_table_h5'),close_notification_func=self.close_notification_func)


class BrokenDevicesNotification(object):
    name = 'Device initialization failed'
    def __init__(self, BLACS):
        # Create the widget
        self._ui = UiLoader().load(os.path.join(PLUGINS_DIR, module, 'broken_device_notification.ui'))

    def get_widget(self):
        return self._ui

    def set_broken_devices(self, device_names):
        self._ui.label.setText('''<html><head/><body><span style=" font-weight:600; color:#ff0000;">BLACS failed to initialize some of your devices.
            It is advised that you solve this problem before using BLACS.
            The devices causing problems were: {}</span></body></html>'''.format(', '.join(device_names)))

    def get_properties(self):
        return {'can_hide':False, 'can_close':False}

    def set_functions(self,show_func,hide_func,close_func,get_state):
        self._show = show_func
        self._hide = hide_func
        self._close = close_func
        self._get_state = get_state

    def get_save_data(self):
        return {}

    def close(self):
        pass

class RecompileNotification(object):
    name = name
    def __init__(self, BLACS):
        # set up the file watching
        self.BLACS = BLACS
        self.filewatcher = None
        self.clean_modified_info = None
        # Create the widget
        self._ui = UiLoader().load(os.path.join(PLUGINS_DIR, module, 'notification.ui'))
        self._ui.button.clicked.connect(self.on_recompile_connection_table)
            
    def get_widget(self):
        return self._ui
        
    def get_properties(self):
        return {'can_hide':True, 'can_close':False}
        
    def set_functions(self, show_func, hide_func, close_func, get_state):
        self._show = show_func
        self._hide = hide_func
        self._close = close_func
        self._get_state = get_state
                
    def on_recompile_connection_table(self,*args,**kwargs):
        self.BLACS['plugins'][module].menu.on_recompile_connection_table()
        
    def callback(self, name, info, event=None):
        if event  == 'deleted':
            logger.info('{} {} ({})'.format(name, event, info))
            inmain(self._show)
        if event  == 'modified':
            logger.info('{} {} ({})'.format(name, event, info))
            inmain(self._show)
        elif event  == 'original':
            logger.info('All watched files restored')
            inmain(self._close)
        elif event == 'restored':
            logger.info('{} {} ({})'.format(name, event, info))
        elif event ==  'debug':
            logger.info(info)

    def setup_filewatching(self, clean_modified_info=None):
        folder_list = []
        file_list = [self.BLACS['connection_table_labscript'], self.BLACS['connection_table_h5file']]
        labconfig = self.BLACS['exp_config']
        try:
            hashable_types = labconfig.get('BLACS/plugins', 'connection_table.hashable_types')
            hashable_types = ast.literal_eval(hashable_types)
        except labconfig.NoOptionError:
            hashable_types = ['.py', '.txt', '.ini', '.json']
        try:
            polling_interval = self.BLACS['exp_config'].getfloat('BLACS/plugins', 'connection_table.polling_interval')
        except labconfig.NoOptionError:
            polling_interval = 1
        logger.info('Using hashable_types: {}; polling_interval: {}'.format(hashable_types, polling_interval))
        
        # append the list of globals
        file_list += self.BLACS['settings'].get_value(Setting,'globals_list')
        # iterate over list, split folders off from files!
        calibration_list = self.BLACS['settings'].get_value(Setting,'calibrations_list')
        for path in calibration_list:
            if os.path.isdir(path):
                folder_list.append(path)
            else:
                file_list.append(path)
        
        # stop watching if we already were
        if self.filewatcher is not None:
            self.filewatcher.stop()
            # Should only be calling this with modified_info not None once at startup:
            assert clean_modified_info is None
            clean_modified_info = self.filewatcher.get_clean_modified_info()
            
        # Start the file watching!
        self.filewatcher = FileWatcher(
            self.callback,
            file_list,
            folder_list,
            clean_modified_info=clean_modified_info,
            hashable_types=hashable_types,
            interval=polling_interval,
        )

    def get_save_data(self):
        if self.clean_modified_info is not None:
            # We are doing a restart after a recompilation - return the modified info
            # that was just saved:
            return {'clean_modified_info': self.clean_modified_info}
        else:
            return {'clean_modified_info': self.filewatcher.get_clean_modified_info()}

    def on_restart(self):
        # Connection table has been sucessfully recompiled, we are restarting. This is
        # the only point, other than when they are first added for watching, that we
        # replace the 'clean' modified info of the files with their current modified
        # info. Since we just recompiled, the current state is the clean state
        self.filewatcher.stop()
        self.clean_modified_info = self.filewatcher.get_modified_info()
        self.filewatcher = None
        # Hide the notification
        self._close()

    def close(self):
        if self.filewatcher is not None:
            self.filewatcher.stop()
    
class Setting(object):
    name = name

    def __init__(self,data):
        # This is our data store!
        self.data = data
        
        self.stores_list = ['globals','calibrations']
        
        for store in self.stores_list:
            if '%s_list'%store not in self.data:
                self.data['%s_list'%store] = []
                
            #set the default sort order if it wasn't previousl saved
            if '%s_sort_order'%store not in self.data:
                self.data['%s_sort_order'%store] = 'ASC'
        
    # Create the page, return the page and an icon to use on the label (the class name attribute will be used for the label text)   
    def create_dialog(self,notebook):
        ui = UiLoader().load(os.path.join(PLUGINS_DIR, module, 'connection_table.ui'))
        
        # Create the models, get the views, and link them!!
        self.models = {}
        self.views = {}
        self.models['globals'] = QStandardItemModel()
        self.models['globals'].setHorizontalHeaderItem(FILEPATH_COLUMN, QStandardItem('Filepath'))
        self.views['globals'] = ui.h5_treeview
        self.views['globals'].setModel(self.models['globals'])
        
        self.models['calibrations'] = QStandardItemModel()
        self.models['calibrations'].setHorizontalHeaderItem(FILEPATH_COLUMN, QStandardItem('Filepath'))
        self.views['calibrations'] = ui.unit_conversion_treeview
        self.views['calibrations'].setModel(self.models['calibrations'])
        
        # Setup the buttons
        ui.add_h5_file.clicked.connect(self.add_global_file)
        ui.delete_h5_file.clicked.connect(self.delete_selected_globals_file)
        ui.add_unitconversion_file.clicked.connect(self.add_calibration_file)
        ui.add_unitconversion_folder.clicked.connect(self.add_calibration_folder)
        ui.delete_unitconversion.clicked.connect(self.delete_selected_conversion_file)
        
        # setup sort indicator changed signals
        self.views['globals'].header().sortIndicatorChanged.connect(self.global_sort_indicator_changed)
        self.views['calibrations'].header().sortIndicatorChanged.connect(self.calibrations_sort_indicator_changed)
        
        #iterate over the two listores
        for store in self.stores_list:
            # If we have saved data in the data store, then load it into the list store
            if '%s_list'%store in self.data:
                for path in self.data['%s_list'%store]:
                    self.models[store].appendRow(QStandardItem(path))
            # otherwise add an empty list to our data store, and leave the liststore empty
            else:
                self.data['%s_list'%store] = []
            
            self.views[store].sortByColumn(FILEPATH_COLUMN,self.order_to_enum(self.data['%s_sort_order'%store]))
        
        return ui,None
    
    def global_sort_indicator_changed(self):
        if 'PySide' in sys.modules.copy():
            if self.views['globals'].header().sortIndicatorOrder() == Qt.SortOrder.AscendingOrder:
                order = 'ASC'
            else:
                order = 'DESC'
        else:
            if self.views['globals'].header().sortIndicatorOrder() == Qt.AscendingOrder:
                order = 'ASC'
            else:
                order = 'DESC'
        self.data['globals_sort_order'] = self.enum_to_order(self.views['globals'].header().sortIndicatorOrder())
        
    def calibrations_sort_indicator_changed(self):
        self.data['calibrations_sort_order'] = self.enum_to_order(self.views['calibrations'].header().sortIndicatorOrder())
    
    def order_to_enum(self, order):
        # if we are accidnetally passed an enum, just return it
        if order not in ['ASC', 'DESC']:
            return order
    
        if 'PySide' in sys.modules.copy():
            if order == 'ASC':
                enum = Qt.SortOrder.AscendingOrder
            else:
                enum = Qt.SortOrder.DescendingOrder
        else:
            if order == 'ASC':
                enum = Qt.AscendingOrder
            else:
                enum = Qt.DescendingOrder
        
        return enum
        
    def enum_to_order(self, enum):
        if 'PySide' in sys.modules.copy():
            if enum == Qt.SortOrder.AscendingOrder:
                order = 'ASC'
            else:
                order = 'DESC'
        else:
            if enum == Qt.AscendingOrder:
                order = 'ASC'
            else:
                order = 'DESC'
        
        return order
    
    def get_value(self,name):
        if name in self.data:
            return self.data[name]
        
        return None
    
    def save(self):
        # transfer the contents of the list store into the data store, and then return the data store
        for store in self.stores_list:
            # clear the existing list
            self.data['%s_list'%store] = []
            for row_index in range(self.models[store].rowCount()):
                self.data['%s_list'%store].append(str(self.models[store].item(row_index).text()))
        
        return self.data
        
    def close(self):
        pass
        
    def add_global_file(self,*args,**kwargs):
        # create file chooser dialog
        dialog = QFileDialog(None,"select globals files", "C:\\", "HDF5 files (*.h5 *.hdf5)")
        dialog.setViewMode(QFileDialog.Detail)
        dialog.setFileMode(QFileDialog.ExistingFiles)
        
        if dialog.exec_():
            selected_files = dialog.selectedFiles()
            for filepath in selected_files:
                filepath = os.path.normpath(filepath)
                # Qt has this weird behaviour where if you type in the name of a file that exists
                # but does not have the extension you have limited the dialog to, the OK button is greyed out
                # but you can hit enter and the file will be selected. 
                # So we must check the extension of each file here!
                if filepath.endswith('.h5') or filepath.endswith('.hdf5'):
                    # make sure the path isn't already in the list
                    if not self.is_filepath_in_store(filepath, 'globals'):
                        self.models['globals'].appendRow(QStandardItem(filepath))
         
            self.views['globals'].sortByColumn(FILEPATH_COLUMN,self.order_to_enum(self.data['globals_sort_order']))
            
        dialog.deleteLater()
            
    def is_filepath_in_store(self,filepath,store):
        for row_index in range(self.models[store].rowCount()):
            if str(filepath) == str(self.models[store].item(row_index).text()):
                return True
        return False
    
    def delete_selected_globals_file(self):
        index_list = self.views['globals'].selectedIndexes()
        while index_list:
            self.models['globals'].takeRow(index_list[0].row())
            index_list = self.views['globals'].selectedIndexes()
        
        self.views['globals'].sortByColumn(FILEPATH_COLUMN,self.order_to_enum(self.data['globals_sort_order']))
            
    def add_calibration_file(self):
        # create file chooser dialog
        dialog = QFileDialog(None,"Select unit conversion scripts", "C:\\", "Python files (*.py *.pyw)")
        dialog.setViewMode(QFileDialog.Detail)
        dialog.setFileMode(QFileDialog.ExistingFiles)
        
        if dialog.exec_():
            selected_files = dialog.selectedFiles()
            for filepath in selected_files:
                filepath = os.path.normpath(filepath)
                # Qt has this weird behaviour where if you type in the name of a file that exists
                # but does not have the extension you have limited the dialog to, the OK button is greyed out
                # but you can hit enter and the file will be selected. 
                # So we must check the extension of each file here!
                if filepath.endswith('.py') or filepath.endswith('.pyw'):
                    # make sure the path isn't already in the list
                    if not self.is_filepath_in_store(filepath,'calibrations'):
                        self.models['calibrations'].appendRow(QStandardItem(filepath))
         
            self.views['calibrations'].sortByColumn(FILEPATH_COLUMN,self.order_to_enum(self.data['calibrations_sort_order']))
        
        dialog.deleteLater()
        
    def add_calibration_folder(self):
        # create file chooser dialog
        dialog = QFileDialog(None,"Select unit conversion folder", "C:\\", "")
        dialog.setViewMode(QFileDialog.Detail)
        dialog.setFileMode(QFileDialog.Directory)
        
        if dialog.exec_():
            selected_files = dialog.selectedFiles()
            for filepath in selected_files:
                filepath = os.path.normpath(filepath)
                # make sure the path isn't already in the list
                if not self.is_filepath_in_store(filepath,'calibrations'):
                    self.models['calibrations'].appendRow(QStandardItem(filepath))
         
            self.views['calibrations'].sortByColumn(FILEPATH_COLUMN,self.order_to_enum(self.data['calibrations_sort_order']))
        
        dialog.deleteLater()
    
    def delete_selected_conversion_file(self):
        index_list = self.views['calibrations'].selectedIndexes()
        while index_list:
            self.models['calibrations'].takeRow(index_list[0].row())
            index_list = self.views['calibrations'].selectedIndexes()
        
        self.views['calibrations'].sortByColumn(FILEPATH_COLUMN,self.order_to_enum(self.data['calibrations_sort_order']))



#####################################################################
#                                                                   #
# __main__.py                                                       #
#                                                                   #
# Copyright 2013, Monash University                                 #
#                                                                   #
# This file is part of the program BLACS, in the labscript suite    #
# (see http://labscriptsuite.org), and is licensed under the        #
# Simplified BSD License. See the license.txt file in the root of   #
# the project for the full license.                                 #
#                                                                   #
#####################################################################
'''BLACS GUI and supporting code
'''
import labscript_utils.excepthook

import os

# Associate app windows with OS menu shortcuts:
import desktop_app
desktop_app.set_process_appid('blacs')


# Splash screen
from labscript_utils.splash import Splash
splash = Splash(os.path.join(os.path.dirname(__file__), 'blacs.svg'))
splash.show()

splash.update_text('importing standard library modules')
import subprocess
import sys
import time
from pathlib import Path
import platform
WINDOWS = platform.system() == 'Windows'

# No splash update for Qt - the splash code has already imported it:
import qtutils
from qtutils import *
import qtutils.icons
from qtutils.qt.QtCore import *
from qtutils.qt.QtGui import *
from qtutils.qt.QtWidgets import *
from qtutils.qt import QT_ENV


splash.update_text("importing zmq and zprocess")
import zmq
import zprocess
from zprocess import raise_exception_in_thread
import zprocess.locking


splash.update_text('importing h5_lock and h5py')
import labscript_utils.h5_lock, h5py


splash.update_text('importing labscript suite modules')
import labscript_utils
from labscript_utils.ls_zprocess import ProcessTree, ZMQServer
from labscript_utils.setup_logging import setup_logging
import labscript_utils.shared_drive
import blacs


process_tree = ProcessTree.instance()
process_tree.zlock_client.set_process_name('BLACS')


# Setup logging
logger = setup_logging('BLACS')
labscript_utils.excepthook.set_logger(logger)

logger.info(f'Python version {sys.version}')
logger.info(f'Platform: {sys.platform}')
logger.info(f'windows version: {sys.getwindowsversion() if WINDOWS else None}')
logger.info(f'PyZMQ version: {zmq.__version__}')
logger.info(f'ZMQ version: {zmq.zmq_version()}')
logger.info(f'h5py version: {h5py.version.info}')
logger.info(f'Qt enviroment: {QT_ENV}')
logger.info(f'PySide/PyQt version: {PYQT_VERSION_STR}')
logger.info(f'Qt version: {QT_VERSION_STR}')
logger.info(f'qtutils version: {qtutils.__version__}')
logger.info(f'zprocess version: {zprocess.__version__}')
logger.info(f'labscript_utils version: {labscript_utils.__version__}')
logger.info(f'BLACS version: {blacs.__version__}')

# Connection Table Code
from labscript_utils.connections import ConnectionTable
#Draggable Tab Widget Code
from labscript_utils.qtwidgets.dragdroptab import DragDropTabWidget
# Lab config code
from labscript_utils.labconfig import LabConfig
from labscript_profile import hostname
# Analysis Submission code
from blacs.analysis_submission import AnalysisSubmission
# Queue Manager Code
from blacs.experiment_queue import QueueManager, QueueTreeview
# Module containing hardware compatibility:
from labscript_utils import device_registry
# Save/restore frontpanel code
from blacs.front_panel_settings import FrontPanelSettings
# Notifications system
from blacs.notifications import Notifications
# Preferences system
from labscript_utils.settings import Settings
#import settings_pages
import blacs.plugins as plugins

from blacs import BLACS_DIR


class BLACSWindow(QMainWindow):

    def closeEvent(self, event):
        if self.blacs.exit_complete:
            event.accept()
            if self.blacs._relaunch:
                logger.info('relaunching BLACS after quit')
                subprocess.Popen([sys.executable] + sys.argv)
        else:
            event.ignore()
            logger.info('destroy called')
            if not self.blacs.exiting:
                self.blacs.exiting = True
                self.blacs.queue.manager_running = False
                self.blacs.settings.close()
                experiment_server.shutdown()
                for module_name, plugin in self.blacs.plugins.items():
                    try:
                        plugin.close()
                    except Exception as e:
                        logger.error('Could not close plugin %s. Error was: %s'%(module_name,str(e)))

                inmain_later(self.blacs.on_save_exit)

            QTimer.singleShot(100,self.close)


class EasterEggButton(QToolButton):
    def __init__(self):
        QToolButton.__init__(self)
        self.setFixedSize(24, 24) # Ensure we're the same size as the other buttons
        self.icon_atom = QIcon(':qtutils/custom/atom')
        self.icon_smiley = QIcon(':qtutils/fugue/smiley-lol')
        self.icon_none = QIcon(None)
        self.icon_mouse_over = self.icon_atom
        self.clicked.connect(self.on_click)

    def enterEvent(self, event):
        """Make the icon only visible on mouse-over"""
        self.setIcon(self.icon_mouse_over)
        return QToolButton.enterEvent(self, event)

    def leaveEvent(self, event):
        if self.icon_mouse_over is self.icon_atom:
            self.setIcon(self.icon_none)
        return QToolButton.leaveEvent(self, event)

    def on_click(self):
        """Run Measure Ball"""
        # Change icon so the user knows something happened, since the game can take a
        # few seconds to start
        self.icon_mouse_over = self.icon_smiley
        self.setIcon(self.icon_mouse_over)
        # Ensure they can't run the game twice at once:
        self.setEnabled(False)
        # Wait for the subprocess in a thread so that we know when it quits:
        qtutils.inthread(self.run_measure_ball)

    def run_measure_ball(self):
        try:
            from subprocess import check_call
            MEASURE_BALL = os.path.join(BLACS_DIR, 'measure_ball', 'RabiBall.exe')
            if not WINDOWS:
                try:
                    check_call(['wine', '--version'])
                except OSError:
                    msg = 'Game cannot be run on Linux or OSX unless WINE is installed'
                    main_window = inmain(self.window)
                    inmain(QMessageBox.warning, main_window, 'BLACS', msg)
                    return
                else:
                    cmd = ['wine', MEASURE_BALL]
            else:
                cmd = [MEASURE_BALL]
            check_call(cmd, cwd=os.path.dirname(MEASURE_BALL))
        finally:
            # Remove smiley, go back to hiding if mouse not over button:
            self.icon_mouse_over = self.icon_atom
            inmain(self.setIcon, self.icon_none)
            inmain(self.setEnabled, True)


class BLACS(object):

    tab_widget_ids = 7

    def __init__(self,application):
        splash.update_text('loading graphical interface')
        self.qt_application = application
        #self.qt_application.aboutToQuit.connect(self.destroy)
        self._relaunch = False
        self.exiting = False
        self.exit_complete = False

        logger.info('Loading BLACS ui')
        #self.ui = BLACSWindow(self).ui
        loader = UiLoader()
        loader.registerCustomWidget(QueueTreeview)
        #loader.registerCustomPromotion('BLACS',BLACSWindow)
        self.ui = loader.load(os.path.join(BLACS_DIR, 'main.ui'), BLACSWindow())
        logger.info('BLACS ui loaded')
        self.ui.blacs=self
        self.tab_widgets = {}
        self.exp_config = exp_config # Global variable
        self.settings_path = settings_path # Global variable
        self.connection_table = connection_table # Global variable
        self.connection_table_h5file = self.exp_config.get('paths','connection_table_h5')
        self.connection_table_labscript = self.exp_config.get('paths','connection_table_py')

        # Setup the UI
        self.ui.main_splitter.setStretchFactor(0,0)
        self.ui.main_splitter.setStretchFactor(1,1)

        self.tablist = {}
        self.panes = {}
        self.settings_dict = {}

        splash.update_text('loading device front panel settings')
        # Find which devices are connected to BLACS, and what their labscript class names are:
        logger.info('finding connected devices in connection table')
        self.attached_devices = self.connection_table.get_attached_devices()

        # Store the panes in a dictionary for easy access
        self.panes['tab_top_vertical_splitter'] = self.ui.tab_top_vertical_splitter
        self.panes['tab_bottom_vertical_splitter'] = self.ui.tab_bottom_vertical_splitter
        self.panes['tab_horizontal_splitter'] = self.ui.tab_horizontal_splitter
        self.panes['main_splitter'] = self.ui.main_splitter

        # Get settings to restore
        logger.info('Loading front panel settings')
        self.front_panel_settings = FrontPanelSettings(self.settings_path, self.connection_table)
        self.front_panel_settings.setup(self)
        settings,question,error,tab_data = self.front_panel_settings.restore()

        # TODO: handle question/error cases

        logger.info('restoring window data')
        self.restore_window(tab_data)

        splash.update_text('creating device tabs...')
        # Create the notebooks
        logger.info('Creating tab widgets')
        for i in range(4):
            self.tab_widgets[i] = DragDropTabWidget(self.tab_widget_ids)
            self.tab_widgets[i].setElideMode(Qt.ElideRight)
            getattr(self.ui,'tab_container_%d'%i).addWidget(self.tab_widgets[i])

        logger.info('Instantiating devices')
        self.failed_device_settings = {}
        for device_name, labscript_device_class_name in list(self.attached_devices.items()):
            try:
                self.settings_dict.setdefault(device_name,{"device_name":device_name})
                # add common keys to settings:
                self.settings_dict[device_name]["connection_table"] = self.connection_table
                self.settings_dict[device_name]["front_panel_settings"] = settings[device_name] if device_name in settings else {}
                self.settings_dict[device_name]["saved_data"] = tab_data[device_name]['data'] if device_name in tab_data else {}
                # Instantiate the device
                logger.info('instantiating %s'%device_name)
                TabClass = device_registry.get_BLACS_tab(labscript_device_class_name)
                self.tablist[device_name] = TabClass(self.tab_widgets[0],self.settings_dict[device_name])
            except Exception:
                self.failed_device_settings[device_name] = {"front_panel": self.settings_dict[device_name]["front_panel_settings"], "save_data": self.settings_dict[device_name]["saved_data"]}
                del self.settings_dict[device_name]
                del self.attached_devices[device_name]
                self.connection_table.remove_device(device_name)
                raise_exception_in_thread(sys.exc_info())

        splash.update_text('instantiating plugins')
        logger.info('Instantiating plugins')
        # setup the plugin system
        settings_pages = []
        self.plugins = {}
        plugin_settings = eval(tab_data['BLACS settings']['plugin_data']) if 'plugin_data' in tab_data['BLACS settings'] else {}
        for module_name, module in plugins.modules.items():
            try:
                # instantiate the plugin
                self.plugins[module_name] = module.Plugin(plugin_settings[module_name] if module_name in plugin_settings else {})
            except Exception:
                logger.exception('Could not instantiate plugin \'%s\'. Skipping' % module_name)

        logger.info('creating plugin tabs')
        # setup the plugin tabs
        for module_name, plugin in self.plugins.items():
            try:
                if hasattr(plugin, 'get_tab_classes'):
                    tab_dict = {}

                    for tab_name, TabClass in plugin.get_tab_classes().items():
                        settings_key = "{}: {}".format(module_name, tab_name)
                        self.settings_dict.setdefault(settings_key, {"tab_name": tab_name})
                        self.settings_dict[settings_key]["front_panel_settings"] = settings[settings_key] if settings_key in settings else {}
                        self.settings_dict[settings_key]["saved_data"] = tab_data[settings_key]['data'] if settings_key in tab_data else {}

                        self.tablist[settings_key] = TabClass(self.tab_widgets[0], self.settings_dict[settings_key])
                        tab_dict[tab_name] = self.tablist[settings_key]

                    if hasattr(plugin, 'tabs_created'):
                        plugin.tabs_created(tab_dict)

            except Exception:
                logger.exception('Could not instantiate tab for plugin \'%s\'. Skipping')

        logger.info('reordering tabs')
        self.order_tabs(tab_data)

        splash.update_text("initialising analysis submission")
        logger.info('starting analysis submission thread')
        # setup analysis submission
        self.analysis_submission = AnalysisSubmission(self,self.ui)
        if 'analysis_data' not in tab_data['BLACS settings']:
            tab_data['BLACS settings']['analysis_data'] = {}
        else:
            tab_data['BLACS settings']['analysis_data'] = eval(tab_data['BLACS settings']['analysis_data'])
        self.analysis_submission.restore_save_data(tab_data['BLACS settings']["analysis_data"])

        splash.update_text("starting queue manager")
        logger.info('starting queue manager thread')
        # Setup the QueueManager
        self.queue = QueueManager(self,self.ui)
        if 'queue_data' not in tab_data['BLACS settings']:
            tab_data['BLACS settings']['queue_data'] = {}
        else:
            # quick fix for qt objects not loading that were saved before qtutil 2 changes
            try:
                tab_data['BLACS settings']['queue_data'] = eval(tab_data['BLACS settings']['queue_data'])
            except NameError:
                tab_data['BLACS settings']['queue_data'] = {}
        self.queue.restore_save_data(tab_data['BLACS settings']['queue_data'])

        blacs_data = {'exp_config':self.exp_config,
                      'ui':self.ui,
                      'set_relaunch':self.set_relaunch,
                      'plugins':self.plugins,
                      'connection_table_h5file':self.connection_table_h5file,
                      'connection_table_labscript':self.connection_table_labscript,
                      'experiment_queue':self.queue
                     }

        def create_menu(parent, menu_parameters):
            if 'name' in menu_parameters:
                if 'menu_items' in menu_parameters:
                    child = parent.addMenu(menu_parameters['name'])
                    for child_menu_params in menu_parameters['menu_items']:
                        create_menu(child,child_menu_params)
                else:
                    if 'icon' in menu_parameters:
                        child = parent.addAction(QIcon(menu_parameters['icon']), menu_parameters['name'])
                    else:
                        child = parent.addAction(menu_parameters['name'])

                if 'action' in menu_parameters:
                    child.triggered.connect(menu_parameters['action'])

            elif 'separator' in menu_parameters:
                parent.addSeparator()

        # setup the Notification system
        logger.info('setting up notification system')
        splash.update_text('setting up notification system')
        self.notifications = Notifications(blacs_data)

        settings_callbacks = []
        for module_name, plugin in self.plugins.items():
            try:
                # Setup settings page
                settings_pages.extend(plugin.get_setting_classes())
                # Setup menu
                if plugin.get_menu_class():
                    # must store a reference or else the methods called when the menu actions are triggered
                    # (contained in this object) will be garbaged collected
                    menu = plugin.get_menu_class()(blacs_data)
                    create_menu(self.ui.menubar,menu.get_menu_items())
                    plugin.set_menu_instance(menu)

                # Setup notifications
                plugin_notifications = {}
                for notification_class in plugin.get_notification_classes():
                    self.notifications.add_notification(notification_class)
                    plugin_notifications[notification_class] = self.notifications.get_instance(notification_class)
                plugin.set_notification_instances(plugin_notifications)

                # Register callbacks
                callbacks = plugin.get_callbacks()
                # save the settings_changed callback in a separate list for setting up later
                if isinstance(callbacks,dict) and 'settings_changed' in callbacks:
                    settings_callbacks.append(callbacks['settings_changed'])

            except Exception:
                logger.exception('Plugin \'%s\' error. Plugin may not be functional.'%module_name)


        # setup the BLACS preferences system
        splash.update_text('setting up preferences system')
        logger.info('setting up preferences system')
        self.settings = Settings(file=self.settings_path, parent = self.ui, page_classes=settings_pages)
        for callback in settings_callbacks:
            self.settings.register_callback(callback)

        # update the blacs_data dictionary with the settings system
        blacs_data['settings'] = self.settings

        for module_name, plugin in self.plugins.items():
            try:
                plugin.plugin_setup_complete(blacs_data)
            except Exception:
                logger.exception('Error in plugin_setup_complete() for plugin \'%s\'. Trying again with old call signature...' % module_name)
                # backwards compatibility for old plugins
                try:
                    plugin.plugin_setup_complete()
                    logger.warning('Plugin \'%s\' using old API. Please update Plugin.plugin_setup_complete method to accept a dictionary of blacs_data as the only argument.'%module_name)
                except Exception:
                    logger.exception('Plugin \'%s\' error. Plugin may not be functional.'%module_name)

        # Connect menu actions
        self.ui.actionOpenPreferences.triggered.connect(self.on_open_preferences)
        self.ui.actionSave.triggered.connect(self.on_save_front_panel)
        self.ui.actionOpen.triggered.connect(self.on_load_front_panel)
        self.ui.actionExit.triggered.connect(self.ui.close)

        # Add hidden easter egg button to a random tab:
        logger.info('hiding easter eggs')
        import random
        if self.tablist:
            random_tab = random.choice(list(self.tablist.values())) 
            self.easter_egg_button = EasterEggButton()
            # Add the button before the other buttons in the tab's header:
            header = random_tab._ui.horizontalLayout
            for i in range(header.count()):
                if isinstance(header.itemAt(i).widget(), QToolButton):
                    header.insertWidget(i, self.easter_egg_button)
                    break

        splash.update_text('done')
        logger.info('showing UI')
        self.ui.show()

    def set_relaunch(self,value):
        self._relaunch = bool(value)

    def restore_window(self,tab_data):
        # read out position settings:
        try:
            # There are some dodgy hacks going on here to try and restore the window position correctly
            # Unfortunately Qt has two ways of measuring teh window position, one with the frame/titlebar
            # and one without. If you use the one that measures including the titlebar, you don't
            # know what the window size was when the window was UNmaximized.
            #
            # Anyway, no idea if this works cross platform (tested on windows 8)
            # Feel free to rewrite this, along with the code in front_panel_settings.py
            # which stores the values
            #
            # Actually this is a waste of time because if you close when maximized, reoopen and then
            # de-maximize, the window moves to a random position (not the position it was at before maximizing)
            # so bleh!
            self.ui.move(
                tab_data['BLACS settings']["window_xpos"]
                - tab_data['BLACS settings']['window_frame_width'] // 2,
                tab_data['BLACS settings']["window_ypos"]
                - tab_data['BLACS settings']['window_frame_height']
                + tab_data['BLACS settings']['window_frame_width'] // 2,
            )
            self.ui.resize(
                tab_data['BLACS settings']["window_width"],
                tab_data['BLACS settings']["window_height"],
            )

            if 'window_maximized' in tab_data['BLACS settings'] and tab_data['BLACS settings']['window_maximized']:
                self.ui.showMaximized()

            for pane_name,pane in self.panes.items():
                pane.setSizes(tab_data['BLACS settings'][pane_name])

        except Exception as e:
            logger.exception("Unable to load window and notebook defaults. Exception:"+str(e))

    def order_tabs(self,tab_data):
        # Move the tabs to the correct notebook
        for tab_name in self.tablist.keys():
            notebook_num = 0
            if tab_name in tab_data:
                notebook_num = int(tab_data[tab_name]["notebook"])
                if notebook_num not in self.tab_widgets:
                    notebook_num = 0

            #Find the notebook the tab is in, and remove it:
            for notebook in self.tab_widgets.values():
                tab_index = notebook.indexOf(self.tablist[tab_name]._ui)
                if tab_index != -1:
                    tab_text = notebook.tabText(tab_index)
                    notebook.removeTab(tab_index)
                    self.tab_widgets[notebook_num].addTab(self.tablist[tab_name]._ui,tab_text)
                    break

        splash.update_text('restoring tab positions...')
        # # Now that all the pages are created, reorder them!
        for tab_name in self.tablist.keys():
            if tab_name in tab_data:
                notebook_num = int(tab_data[tab_name]["notebook"])
                if notebook_num in self.tab_widgets:
                    self.tab_widgets[notebook_num].tab_bar.moveTab(self.tab_widgets[notebook_num].indexOf(self.tablist[tab_name]._ui),int(tab_data[tab_name]["page"]))

        # # Now that they are in the correct order, set the correct one visible
        for tab_name, data in tab_data.items():
            if tab_name == 'BLACS settings':
                continue
            # if the notebook still exists and we are on the entry that is visible
            if bool(data["visible"]) and int(data["notebook"]) in self.tab_widgets:
                self.tab_widgets[int(data["notebook"])].tab_bar.setCurrentIndex(int(data["page"]))

    def update_all_tab_settings(self,settings,tab_data):
        for tab_name,tab in self.tablist.items():
            self.settings_dict[tab_name]["front_panel_settings"] = settings[tab_name] if tab_name in settings else {}
            self.settings_dict[tab_name]["saved_data"] = tab_data[tab_name]['data'] if tab_name in tab_data else {}
            tab.update_from_settings(self.settings_dict[tab_name])


    def on_load_front_panel(self,*args,**kwargs):
        # get the file:
        # create file chooser dialog
        dialog = QFileDialog(None,"Select file to load", self.exp_config.get('paths','experiment_shot_storage'), "HDF5 files (*.h5 *.hdf5)")
        dialog.setViewMode(QFileDialog.Detail)
        dialog.setFileMode(QFileDialog.ExistingFile)
        if dialog.exec_():
            selected_files = dialog.selectedFiles()
            filepath = str(selected_files[0])
            # Qt has this weird behaviour where if you type in the name of a file that exists
            # but does not have the extension you have limited the dialog to, the OK button is greyed out
            # but you can hit enter and the file will be selected.
            # So we must check the extension of each file here!
            if filepath.endswith('.h5') or filepath.endswith('.hdf5'):
                try:
                    # TODO: Warn that this will restore values, but not channels that are locked
                    message = QMessageBox()
                    message.setText("""Warning: This will modify front panel values and cause device output values to update.
                    \nThe queue and files waiting to be sent for analysis will be cleared.
                    \n
                    \nNote: Channels that are locked will not be updated.\n\nDo you wish to continue?""")
                    message.setIcon(QMessageBox.Warning)
                    message.setWindowTitle("BLACS")
                    message.setStandardButtons(QMessageBox.Yes|QMessageBox.No)

                    if message.exec_() == QMessageBox.Yes:
                        front_panel_settings = FrontPanelSettings(filepath, self.connection_table)
                        settings,question,error,tab_data = front_panel_settings.restore()
                        #TODO: handle question/error

                        # Restore window data
                        self.restore_window(tab_data)
                        self.order_tabs(tab_data)
                        self.update_all_tab_settings(settings,tab_data)

                        # restore queue data
                        if 'queue_data' not in tab_data['BLACS settings']:
                            tab_data['BLACS settings']['queue_data'] = {}
                        else:
                            # quick fix for qt objects not loading that were saved before qtutil 2 changes
                            try:
                                tab_data['BLACS settings']['queue_data'] = eval(tab_data['BLACS settings']['queue_data'])
                            except NameError:
                                tab_data['BLACS settings']['queue_data'] = {}
                        self.queue.restore_save_data(tab_data['BLACS settings']['queue_data'])
                        # restore analysis data
                        if 'analysis_data' not in tab_data['BLACS settings']:
                            tab_data['BLACS settings']['analysis_data'] = {}
                        else:
                            tab_data['BLACS settings']['analysis_data'] = eval(tab_data['BLACS settings']['analysis_data'])
                        self.analysis_submission.restore_save_data(tab_data['BLACS settings']["analysis_data"])
                except Exception as e:
                    logger.exception("Unable to load the front panel in %s."%(filepath))
                    message = QMessageBox()
                    message.setText("Unable to load the front panel. The error encountered is printed below.\n\n%s"%str(e))
                    message.setIcon(QMessageBox.Information)
                    message.setWindowTitle("BLACS")
                    message.exec_()
                finally:
                    dialog.deleteLater()
            else:
                dialog.deleteLater()
                message = QMessageBox()
                message.setText("You did not select a file ending with .h5 or .hdf5. Please try again")
                message.setIcon(QMessageBox.Information)
                message.setWindowTitle("BLACS")
                message.exec_()
                QTimer.singleShot(10,self.on_load_front_panel)

    def on_save_exit(self):
        # Save front panel
        data = self.front_panel_settings.get_save_data()

        if len(self.failed_device_settings) > 0:
            message = ('Save data from broken tabs? \n Broken tabs are: \n {}'.format(list(self.failed_device_settings.keys())))
            reply = QMessageBox.question(self.ui, 'Save broken tab data?', message,
                                               QMessageBox.Yes | QMessageBox.No)
            if reply == QMessageBox.Yes:
                data[0].update(self.failed_device_settings)

        # with h5py.File(self.settings_path,'r+') as h5file:
           # if 'connection table' in h5file:
               # del h5file['connection table']

        self.front_panel_settings.save_front_panel_to_h5(self.settings_path,data[0],data[1],data[2],data[3],{"overwrite":True},force_new_conn_table=True)
        logger.info('Shutting down workers')
        for tab in self.tablist.values():
            # Tell tab to shutdown its workers if it has a method to do so.
            if hasattr(tab, 'shutdown_workers'):
                tab.shutdown_workers()

        QTimer.singleShot(100, self.finalise_quit)

    def finalise_quit(self, deadline=None, pending_threads=None):
        logger.info('finalise_quit called')
        WORKER_SHUTDOWN_TIMEOUT = 2
        if deadline is None:
            deadline = time.time() + WORKER_SHUTDOWN_TIMEOUT
        if pending_threads is None:
            pending_threads = {}
        overdue = time.time() > deadline
        # Check for worker shutdown completion:
        for name, tab in list(self.tablist.items()):
            # Immediately close tabs that don't support finalise_close_tab()
            if not hasattr(tab, 'finalise_close_tab'):
                try:
                    current_page = tab.close_tab(finalise=False)
                except Exception as e:
                    logger.error('Couldn\'t close tab:\n%s' % str(e))
                del self.tablist[name]
                continue
            fatal_error = tab.state == 'fatal error'
            if not tab.shutdown_workers_complete and overdue or fatal_error:
                # Give up on cleanly shutting down this tab's worker processes:
                tab.shutdown_workers_complete = True
            if tab.shutdown_workers_complete:
                if name not in pending_threads:
                    # Either worker shutdown completed or we gave up. Close the tab.
                    try:
                        current_page = tab.close_tab(finalise=False)
                    except Exception as e:
                        logger.error('Couldn\'t close tab:\n%s' % str(e))
                        del self.tablist[name]
                    else:
                        # Call finalise_close_tab in a thread since it can be blocking.
                        # It has its own timeout however, so we do not need to keep
                        # track of whether tabs are taking too long to finalise_close()
                        pending_threads[name] = inthread(
                            tab.finalise_close_tab, current_page
                        )
                elif not pending_threads[name].is_alive():
                    # finalise_close_tab completed, tab is closed and worker terminated
                    pending_threads[name].join()
                    del pending_threads[name]
                    del self.tablist[name]
        if not self.tablist:
            # All tabs are closed.
            self.exit_complete = True
            logger.info('quitting')
            return
        QTimer.singleShot(100, lambda: self.finalise_quit(deadline, pending_threads))
            

    def on_save_front_panel(self,*args,**kwargs):
        data = self.front_panel_settings.get_save_data()

        # Open save As dialog
        dialog = QFileDialog(None,"Save BLACS state", self.exp_config.get('paths','experiment_shot_storage'), "HDF5 files (*.h5)")
        try:
            dialog.setViewMode(QFileDialog.Detail)
            dialog.setFileMode(QFileDialog.AnyFile)
            dialog.setAcceptMode(QFileDialog.AcceptSave)

            if dialog.exec_():
                current_file = str(dialog.selectedFiles()[0])
                if not current_file.endswith('.h5'):
                    current_file += '.h5'
                self.front_panel_settings.save_front_panel_to_h5(current_file,data[0],data[1],data[2],data[3])
        except Exception:
            raise
        finally:
            dialog.deleteLater()

    def on_open_preferences(self,*args,**kwargs):
        self.settings.create_dialog()

class ExperimentServer(ZMQServer):
    def handler(self, h5_filepath):
        print(h5_filepath)
        message = self.process(h5_filepath)
        logger.info('Request handler: %s ' % message.strip())
        return message

    @inmain_decorator(wait_for_return=True)
    def process(self,h5_filepath):
        # Convert path to local slashes and shared drive prefix:
        logger.info('received filepath: %s'%h5_filepath)
        h5_filepath = labscript_utils.shared_drive.path_to_local(h5_filepath)
        logger.info('local filepath: %s'%h5_filepath)
        return app.queue.process_request(h5_filepath)


if __name__ == '__main__':
    if 'tracelog' in sys.argv:
        ##########
        import labscript_utils.tracelog
        labscript_utils.tracelog.log(os.path.join(BLACS_DIR, 'blacs_trace.log'),
                                     ['__main__','BLACS.tab_base_classes',
                                      'qtutils',
                                      'labscript_utils.qtwidgets.ddsoutput',
                                      'labscript_utils.qtwidgets.analogoutput',
                                      'BLACS.hardware_interfaces.ni_pcie_6363',
                                      'BLACS.hardware_interfaces.output_classes',
                                      'BLACS.device_base_class',
                                      'BLACS.tab_base_classes',
                                      'BLACS.plugins.connection_table',
                                      'BLACS.recompile_and_restart',
                                      'filewatcher',
                                      'queue',
                                      'notifications',
                                      'connections',
                                      'analysis_submission',
                                      'settings',
                                      'front_panel_settings',
                                      'labscript_utils.h5_lock',
                                      'labscript_utils.shared_drive',
                                      'labscript_utils.labconfig',
                                      'zprocess',
                                     ], sub=True)

    splash.update_text('loading labconfig')
    required_config_params = {
        "DEFAULT": ["apparatus_name", "app_saved_configs"],
        "programs": ["text_editor", "text_editor_arguments",],
        "paths": ["shared_drive", "connection_table_h5", "connection_table_py",],
        "ports": ["BLACS", "lyse"],
    }
    exp_config = LabConfig(required_params=required_config_params)
    settings_dir = Path(exp_config.get('DEFAULT', 'app_saved_configs'), 'blacs')
    if not settings_dir.exists():
        os.makedirs(settings_dir, exist_ok=True)
    settings_path = str(settings_dir / f'{hostname()}_BLACS.h5')

    port = int(exp_config.get('ports','BLACS'))

    # Start experiment server
    splash.update_text('starting experiment server')
    experiment_server = ExperimentServer(port)

    # Create Connection Table object
    splash.update_text('loading connection table')
    logger.info('About to load connection table: %s'%exp_config.get('paths','connection_table_h5'))
    connection_table_h5_file = exp_config.get('paths','connection_table_h5')
    connection_table = ConnectionTable(connection_table_h5_file, logging_prefix='BLACS', exceptions_in_thread=True)

    logger.info('connection table loaded')

    splash.update_text('initialising Qt application')
    qapplication = QApplication.instance()
    if qapplication is None:
        qapplication = QApplication(sys.argv)
    qapplication.setAttribute(Qt.AA_DontShowIconsInMenus, False)
    logger.info('QApplication instantiated')
    app = BLACS(qapplication)

    logger.info('BLACS instantiated')
    splash.hide()

    def execute_program():
        qapplication.exec_()

    sys.exit(execute_program())



#####################################################################
#                                                                   #
# /plugins/memory/__init__.py                                       #
#                                                                   #
# Copyright 2013, Monash University                                 #
#                                                                   #
# This file is part of the program BLACS, in the labscript suite    #
# (see http://labscriptsuite.org), and is licensed under the        #
# Simplified BSD License. See the license.txt file in the root of   #
# the project for the full license.                                 #
#                                                                   #
#####################################################################
import os
import logging
import gc
from blacs import BLACS_DIR
from labscript_utils import memprof

FILEPATH_COLUMN = 0
name = "Memory Profile"
module = "memory" # should be folder name
logger = logging.getLogger('BLACS.plugin.%s'%module)

class Plugin(object):
    def __init__(self,initial_settings):
        self.menu = None
        self.notifications = {}
        self.initial_settings = initial_settings
        self.BLACS = None
        
    def get_menu_class(self):
        return Menu
        
    def get_notification_classes(self):
        return []
        
    def get_setting_classes(self):
        return []
        
    def get_callbacks(self):
        {}
        
    def set_menu_instance(self,menu):
        self.menu = menu
        
    def set_notification_instances(self,notifications):
        pass
        
    def plugin_setup_complete(self, BLACS):
        self.BLACS = BLACS
        
    def get_save_data(self):
        return {}
    
    def close(self):
        pass

class Menu(object):
    def __init__(self,BLACS):
        self.BLACS = BLACS
        self.close_notification_func = None
        memprof.start(filepath=os.path.join(BLACS_DIR, 'memprof.txt'))        
        
    def get_menu_items(self):
        return {'name':name,        
                'menu_items':[{'name':'Garbage collect',
                               'action':gc.collect,
                               'icon': ':/qtutils/fugue/memory'
                              },
                              {'name':'Reset profiler',
                               'action':memprof.start,
                               'icon': ':/qtutils/fugue/counter-reset'
                              },
                              {'name':'Diff memory usage',
                               'action':memprof.check,
                               'icon': ':/qtutils/fugue/tables'
                              }
                             ]                                
               }
    



#####################################################################
#                                                                   #
# /plugins/theme/__init__.py                                        #
#                                                                   #
# Copyright 2013, Monash University                                 #
#                                                                   #
# This file is part of the program BLACS, in the labscript suite    #
# (see http://labscriptsuite.org), and is licensed under the        #
# Simplified BSD License. See the license.txt file in the root of   #
# the project for the full license.                                 #
#                                                                   #
#####################################################################
import logging
import os

from qtutils import *

from blacs.plugins import PLUGINS_DIR

name = "GUI Theme"
module = "theme" # should be folder name
logger = logging.getLogger('BLACS.plugin.%s'%module)


DEFAULT_STYLESHEET = """DigitalOutput {
    font-size: 12px;
    background-color: rgb(50,100,50,255);
    border: 1px solid rgb(50,100,50,128);
    border-radius: 3px;
    padding: 2px;
    color: #202020;
}

DigitalOutput:hover {
    background-color: rgb(50,130,50);
    border: None;
}

DigitalOutput:disabled{
   background-color: rgb(50,100,50,128);
   color: #505050;
}

DigitalOutput:checked {
    background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                                      stop: 0 rgb(32,200,32), stop: 1 rgb(32,255,32));
    border: 1px solid #8f8f91;
    color: #000000;
}

DigitalOutput:hover:checked {
    background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                                      stop: 0 rgb(32,200,32), stop: 1 rgb(120,255,120));
    border: 1px solid #8f8f91;
}

DigitalOutput:checked:disabled{
   background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                                      stop: 0 rgba(32,200,32,128), stop: 1 rgba(32,255,32,128));
   color: #606060;
}

InvertedDigitalOutput {
    font-size: 12px;
    background-color: rgb(70,100,170,255);
    border: 1px solid rgb(70,100,170,128);
    border-radius: 3px;
    padding: 2px;
    color: #202020;
}

InvertedDigitalOutput:hover {
    background-color: rgb(70, 130, 220);
    border: None;
}

InvertedDigitalOutput:disabled{
   background-color: rgba(70,100,170,128);
   color: #505050;
}

InvertedDigitalOutput:checked {
    background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                                      stop: 0 rgb(50,150,221), stop: 1 rgb(32,192,255));
    border: 1px solid #8f8f91;
    color: #000000;
}

InvertedDigitalOutput:hover:checked {
    background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                                      stop: 0 rgb(50,150,221), stop: 1 rgb(120,192,255));
    border: 1px solid #8f8f91;
}

InvertedDigitalOutput:checked:disabled{
   background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                                      stop: 0 rgba(50,150,221,128), stop: 1 rgba(32,192,255,128));
   color: #606060;
}
 """


def is_default_stylesheet(stylesheet):
    """Return whether a stylesheet is the same as the default stylesheet, modulo whitespace"""

    def no_whitespace(s):
        return "".join(s.split())

    return no_whitespace(str(stylesheet)) == no_whitespace(DEFAULT_STYLESHEET) 


class Plugin(object):
    def __init__(self,initial_settings):
        self.menu = None
        self.notifications = {}
        self.BLACS = None
        self.initial_settings = initial_settings
        
    def get_menu_class(self):
        return None
        
    def get_notification_classes(self):
        return []
        
    def get_setting_classes(self):
        return [Setting]
        
    def get_callbacks(self):
        return {'settings_changed':self.update_stylesheet}
        
    def update_stylesheet(self):
        if self.BLACS is not None:
            # show centralwidget as a workaround to fix stylsheets
            # not beeing applied under PyQt5 on first draw
            self.BLACS['ui'].centralwidget.show()
            stylesheet_settings = self.BLACS['settings'].get_value(Setting,"stylesheet")
            self.BLACS['ui'].centralwidget.setStyleSheet(self.unmodified_stylesheet + stylesheet_settings)
        
    def set_menu_instance(self,menu):
        self.menu = menu
                
    def set_notification_instances(self,notifications):
        self.notifications = notifications
        
    def plugin_setup_complete(self, BLACS):
        self.BLACS = BLACS
        self.unmodified_stylesheet = self.BLACS['ui'].centralwidget.styleSheet()
        self.update_stylesheet()
    
    def get_save_data(self):
        return {}
    
    def close(self):
        pass
        
    
class Setting(object):
    name = name

    def __init__(self,data):
        # This is our data store!
        self.data = data
        
        if 'stylesheet' not in self.data or not self.data['stylesheet']:
            # If it's absent or an empty string, use the default stylesheet:
            self.data['stylesheet'] = DEFAULT_STYLESHEET
    
    def on_set_green_button_theme(self):
        self.widgets['stylesheet'].appendPlainText(DEFAULT_STYLESHEET)
        
    # Create the page, return the page and an icon to use on the label (the class name attribute will be used for the label text)   
    def create_dialog(self,notebook):
        ui = UiLoader().load(os.path.join(PLUGINS_DIR, module, 'theme.ui'))
        
        # restore current stylesheet
        ui.stylesheet_text.setPlainText(self.data['stylesheet'])
        ui.example_button.clicked.connect(self.on_set_green_button_theme)
        
        # save reference to widget
        self.widgets = {}
        self.widgets['stylesheet'] = ui.stylesheet_text
        self.widgets['example_button'] = ui.example_button
        
        return ui,None
    
    def get_value(self,name):
        if name in self.data:
            return self.data[name]
        
        return None
    
    def save(self):
        stylesheet = str(self.widgets['stylesheet'].toPlainText())
        if not stylesheet.endswith('\n'):
            # This is a way to distinguish between an intentionally blank
            # stylesheet, and an empty string, which used to be what was
            # stored When the user had made no changes, which now we take to
            # imply that they want to use the default stylesheet:
            stylesheet += '\n'
        self.data['stylesheet'] = stylesheet
        data = self.data.copy()
        if is_default_stylesheet(stylesheet):
            # Only save if it is not the default stylesheet:
            del data['stylesheet']
        return data
        
    def close(self):
        self.widgets['example_button'].clicked.disconnect(self.on_set_green_button_theme)
        
    



#####################################################################
#                                                                   #
# /plugins/cycle_time/__init__.py                                   #
#                                                                   #
# Copyright 2017, Monash Univiersity and contributors               #
#                                                                   #
# This file is part of the program BLACS, in the labscript suite    #
# (see http://labscriptsuite.org), and is licensed under the        #
# Simplified BSD License. See the license.txt file in the root of   #
# the project for the full license.                                 #
#                                                                   #
#####################################################################
from queue import Queue, Empty
from time import monotonic
import time
import logging
import os

from qtutils import UiLoader, inmain, inmain_decorator

import labscript_utils.h5_lock
import h5py

from qtutils.qt.QtGui import QIcon
from qtutils.qt.QtCore import QSize

from blacs.plugins import PLUGINS_DIR, callback
from labscript_utils.properties import get_attribute

name = "cycle_time"
module = "cycle_time" # should be folder name
logger = logging.getLogger('BLACS.plugin.%s'%module)


class Plugin(object):
    def __init__(self, initial_settings):
        self.menu = None
        self.notifications = {}
        self.initial_settings = initial_settings
        self.BLACS = None
        self.time_of_last_shot = None
        self.queue = Queue()
        self.target_cycle_time = None
        self.delay_after_programming = None
        self.next_target_cycle_time = None
        self.next_delay_after_programming = None

    def plugin_setup_complete(self, BLACS):
        self.BLACS = BLACS
        self.queue_manager = self.BLACS['experiment_queue']

    def get_save_data(self):
        return {}
    
    def get_callbacks(self):
        return {
            'pre_transition_to_buffered': self.pre_transition_to_buffered,
            'science_starting': self.science_starting,
        }
        
    def _abort(self):
        self.queue.put('abort')

    @callback(priority=100) # this callback should run after all other callbacks.
    def pre_transition_to_buffered(self, h5_filepath):
        # Delay according to the settings of the previously run shot, and save the
        # settings for the upcoming shot:
        self.target_cycle_time = self.next_target_cycle_time
        self.delay_after_programming = self.next_delay_after_programming

        with h5py.File(h5_filepath, 'r') as f:
            try:
                group = f['shot_properties']
            except KeyError:
                # Nothing for us to do
                return
            self.next_target_cycle_time = get_attribute(group, 'target_cycle_time')
            self.next_delay_after_programming = get_attribute(
                group, 'cycle_time_delay_after_programming'
            )

        if not self.delay_after_programming:
            self.do_delay(h5_filepath)

    # this callback should run before the progress bar plugin, but after all other
    # callbacks. The priority should be set accordingly.
    @callback(priority=100) 
    def science_starting(self, h5_filepath):
        if self.delay_after_programming:
            self.do_delay(h5_filepath)

    def do_delay(self, h5_filepath):
        if self.target_cycle_time is not None and self.time_of_last_shot is not None:
            # Wait until it has been self.target_cycle_time since the start of the last
            # shot. Otherwise, return immediately.
            deadline = self.time_of_last_shot + self.target_cycle_time
            inmain(self.BLACS['ui'].queue_abort_button.clicked.connect, self._abort)
            # Store the current queue manager status, to restore it after we are done:
            previous_status = self.queue_manager.get_status()
            while True:
                remaining = deadline - monotonic()
                if remaining <= 0:
                    break
                self.queue_manager.set_status(
                    'Waiting {:.1f}s for target cycle time'.format(remaining),
                    h5_filepath,
                )
                try:
                    self.queue.get(timeout=remaining % 0.1)
                    break # Got an abort
                except Empty:
                    continue
            # Disconnect from the abort button:
            inmain(self.BLACS['ui'].queue_abort_button.clicked.disconnect, self._abort)
            # Restore previous_status:
            self.queue_manager.set_status(previous_status, h5_filepath)

        self.time_of_last_shot = monotonic()

    # The rest of these are boilerplate:
    def close(self):
        pass
    
    def get_menu_class(self):
        return None
        
    def get_notification_classes(self):
        return []
        
    def get_setting_classes(self):
        return []
    
    def set_menu_instance(self, menu):
        self.menu = menu
        
    def set_notification_instances(self, notifications):
        self.notifications = notifications
        
    
